Package: rhdf5client
Title: Access HDF5 content from HDF Scalable Data Service
Description: This package provides functionality for reading data from HDF Scalable Data Service
	from within R.  The HSDSArray function bridges from HSDS to the user via the DelayedArray
	interface.  Bioconductor manages an open HSDS instance graciously provided by John
	Readey of the HDF Group.
Version: 1.34.2
Authors@R: c(
 person("Samuela", "Pollack", role = c("aut"),email = "spollack@jimmy.harvard.edu"), 
 person("Shweta", "Gopaulakrishnan", role = c("aut"), email = "reshg@channing.harvard.edu"),
 person("BJ", "Stubbs", role = c("aut"), email = "rebjh@channing.harvard.edu"),
 person("Alexey", "Sergushichev", role = c("aut", "cre"), email = "alsergbox@gmail.com"),
 person("Vincent", "Carey", role = c("aut"), email = "stvjc@channing.harvard.edu")
 )
Suggests: knitr, testthat, BiocStyle, DT, rmarkdown
Imports: httr, rjson, utils, data.table
Depends: R (>= 3.6), methods, DelayedArray
License: Artistic-2.0
LazyLoad: yes
Collate: URL_hsds.R indx.R 'rhdf5client2.R' 'Source.R' 'File.R'
        'Dataset.R' 'RHDF5Array.R'
VignetteBuilder: knitr
Encoding: UTF-8
biocViews: DataImport, Software, Infrastructure
RoxygenNote: 7.3.3
git_url: https://git.bioconductor.org/packages/rhdf5client
git_branch: RELEASE_3_23
git_last_commit: 0ed5c74
git_last_commit_date: 2026-06-01
Repository: Bioconductor 3.23
Date/Publication: 2026-06-02
NeedsCompilation: yes
Packaged: 2026-06-02 23:55:44 UTC; biocbuild
Author: Samuela Pollack [aut],
  Shweta Gopaulakrishnan [aut],
  BJ Stubbs [aut],
  Alexey Sergushichev [aut, cre],
  Vincent Carey [aut]
Maintainer: Alexey Sergushichev <alsergbox@gmail.com>

# Generated by roxygen2: do not edit by hand

export(HSDSArray)
export(HSDSDataset)
export(HSDSFile)
export(HSDSSource)
export(URL_hsds)
export(check_hsds)
export(extractCompoundJSON)
export(getData)
export(getHSDSFileList)
export(isplit)
export(listDatasets)
export(listDomains)
export(sproc)
exportClasses(HSDSArray)
exportClasses(HSDSMatrix)
import(DelayedArray)
import(httr)
import(methods)
import(rjson)
importFrom(DelayedArray,new_DelayedArray)
importFrom(data.table,as.data.table)
importFrom(data.table,rbindlist)
importFrom(utils,strcapture)
useDynLib(rhdf5client, .registration = TRUE)

deprecate_msg = paste0("This function is deprecated. The new interface to rhdf5client",
 " is exclusively through its DelayedArray backend HSDSArray")

#' An S4 class to represent an HDF5 file accessible from a server.
#'
#' @slot HSDSSource an object of type HSDSSource
#' @slot domain the file's domain on the server; more or less, an alias for its 
#' location in the external server file system
#' @slot dsetdf a data.frame that caches often-used information about the file
#' @slot type string representing type of the file: "domain" for file and "folder" for directory
setClass("HSDSFile", representation(src="HSDSSource", domain="character", dsetdf="data.frame", type="character"))

#' Construct an object of type HSDSFile
#'
#' A HSDSFile is a representation of an HDF5 file the contents of which are accessible 
#' exposed by a HDF5 server. 
#'
#' This function is deprecated and will be defunct in the next release.
#'
#' @name HSDSFile
#' @param src an object of type HSDSSource, the server which exposes the file
#' @param domain the domain string; the file's location on the server's
#' file system.
#' @return an initialized object of type HSDSFile
#' @examples
#' if (check_hsds()) {
#'  src <- HSDSSource(URL_hsds())
#'  f10x <- HSDSFile(src, '/shared/bioconductor/patelGBMSC.h5')
#' }
#' @export
HSDSFile <- function(src, domain)  {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  request <- paste0(src@endpoint, '?domain=', domain)
  response <- submitRequest(request)

  if (is.null(request))  {
    warning("no such file")
    return(NULL)
  }
  type <- response$class
  
  if (type == "domain") {
    dsetdf <- findDatasets(src, domain)  
  } else {
    dsetdf <- data.frame(paths=c(), uuids = c(), stringsAsFactors = FALSE)
  }
  
  
  obj <- new("HSDSFile", src=src, domain=domain, dsetdf=dsetdf, type=type)
}

.HSDSFile <- function(src, domain)  {  # after deprecation cycle this private function will be used
  request <- paste0(src@endpoint, '?domain=', domain)
  response <- tryCatch(
    submitRequest(request),
    error=function(e) { NULL }
  )  
  if (is.null(request))  {
    warning("no such file")
    return(NULL)
  }
  type <- response$class
  
  if (type == "domain") {
    dsetdf <- findDatasets(src, domain)  
  } else {
    dsetdf <- data.frame(paths=c(), uuids = c(), stringsAsFactors = FALSE)
  }
  
  obj <- new("HSDSFile", src=src, domain=domain, dsetdf=dsetdf, type=type)
}

#' Search inner file hierarchy for datasets
#' 
#' The datasets in an HDF5 file are organized internally by groups.
#' This routine traverses the internal group hiearchy, locates
#' all datasets and prints a list of them. Note that if the 
#' file's group hiearchy is complex, this could be time-consuming.
#'
#' This function is deprecated and will be defunct in the next release.
#'
#' @param file an object of type HSDSFile to be searched
#' 
#' @return a list of inner-paths 
#' 
#' @examples
#' if (check_hsds()) {
#'  src <- HSDSSource(URL_hsds())
#'  f <- HSDSFile(src, '/shared/bioconductor/patelGBMSC.h5')
#'  listDatasets(f)
#' }
#' @export
listDatasets <- function(file)  {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  file@dsetdf[['paths']]
}

#  private - traverse internal file hiearchy, find datasets, and
#  cache often-accessed information in a data.frame for the HSDSFile object.
findDatasets <- function(src, domain)  {

  result <- tryCatch({
    request <- paste0(src@endpoint, '?domain=', domain)
    response <- submitRequest(request)
    fileroot <- response$root

    # ye olde depth-first search
    eee <- new.env(parent=emptyenv())
    eee$results <- c()            # paths to datasets
    eee$uuids <- c()

    search <- function(uuid, path, ee)  {
      # ee$results <- c(ee$results, path)
      request <- paste0(src@endpoint, '/groups/', uuid, '/links?domain=', domain)
      response <- submitRequest(request)
      for (link in response[['links']])  {
        if ('collection' %in% names(link) && link[['collection']] == 'groups')  {
          nxtuuid <- link[['id']]
          nxtpath <- paste0(path, '/', link[['title']])
          search(nxtuuid, nxtpath, ee)
        } else if ('collection' %in% names(link) && link[['collection']] == 'datasets')  {
          nxtuuid <- link[['id']]
          nxtpath <- paste0(path, '/', link[['title']])
          ee$results <- c(ee$results, nxtpath)
          ee$uuids <- c(ee$uuids, nxtuuid)
        }
      }
    }
    search(fileroot, '', eee)
    1
  }, error = function(e) { -1 })

  if (result == -1)  {
    warning(paste0("no datasets for file ", domain), call. = FALSE)
    return(data.frame(paths=c(), uuids = c(), stringsAsFactors = FALSE))
  }
  return(data.frame(paths=eee$results, uuids=eee$uuids, stringsAsFactors = FALSE))
}

setMethod("show", "HSDSFile", function(object) {
 cat(paste("rhdf5client HSDSFile instance from source", object@src@endpoint, "\n"))
 cat(paste("  domain: ", object@domain, "\n"))
 cat("  use listDatasets(...) and HSDSDataset(..., [dataset name]) for more content.\n")
})

#setMethod("show", "HSDSDataset", function(object) {
# cat(paste("rhdf5client HSDSDataset instance, with shape "))
# dput(object@shape)
# cat("  use getData(...) or square brackets to retrieve content.\n")
#})

# private function to get files on a directory on HSDS server
listDomainFiles <- function(src, directory) {
  request <- paste0(src@endpoint, "/domains?domain=", directory)
  response <- submitRequest(request)
  domains <- response[["domains"]]
  
  res <- lapply(domains, function(d) {
    if (d$class == "domain") {
      return(d$name)
    }
    
    stopifnot(d$class == "folder")
    return(c(d$name,
             listDomainFiles(src, d$name)))
  })
  res <- unlist(res)    
  res
}

#' List files on a directory on HSDS-server 
#' 
#' This funstions returns a list of all HDF5-files on the server or all files of the collection
#'
#' @param url, containing url of the server and root domain.
#' @param directory, containing name of the directory
#' 
#' @return List of all HDF5-files on the server or all files of the collection
#' 
#' @examples
#' if (check_hsds()) {
#'  url <- 'https://alserglab.wustl.edu/hsds/?domain=/counts'
#'  getHSDSFileList(url)
#' }
#' @export
getHSDSFileList <- function(url, directory = NULL) {
  src <- httr::parse_url(url)
  root <- src$query$domain
  src <- paste0(src$scheme,'://',src$hostname,'/',src$path)
  src <- HSDSSource(src)
  if (is.null(directory)) {
    directory <- root
  } else {
    directory <- paste0(root, "/", directory)
  }
  
  hdf5FileList  <- listDomainFiles(src, directory)
  hdf5FileList <- grep(".*\\.h5$", hdf5FileList, value = TRUE)
  return(hdf5FileList)
}

library(rhdf5client)
context("indexing infrastructure")

test_that("sproc/isplit work", {
 if (check_hsds()) {
    expect_true(length(isplit(c(1,2,3,4,5,10,15,20,30:40)))==3)

    ii = isplit(c(1,2,3,4,5,10,15,20,30:40))
    ss = structure(c("0:5:1", "9:20:5", "29:40:1"), .Names = c("1", 
   "2", "3"))

    expect_true(identical(ss, unlist(sproc(ii))))
    ii = isplit(c(1:10, seq(50,25,-5), seq(80,100,2)))
    ss = structure(c("0:10:1", "50:24:-5", "79:100:2"), 
      .Names = c("1", "2", "3"))
    expect_true(identical(ss, unlist(sproc(ii))))
    ii = isplit(c(1, 3, 5, 200000, 300000))
    ss = structure(c("0:5:2", "199999:300000:100000"), 
      .Names = c("1", "2"))
    expect_true(identical(ss, unlist(sproc(ii))))
    }
  else TRUE
})

context("HSDSSource")
test_that("Server found", {
 if (check_hsds()) {
  src.hsds <- HSDSSource('https://alserglab.wustl.edu/hsds/')
  doms <- listDomains(src.hsds, '/shared/bioconductor')
  expect_true('/shared/bioconductor/patelGBMSC.h5' %in% doms) 
  # catch exception: non-existent source
  src.fake <- HSDSSource('https://alserglab-nonex.wustl.edu/hsds/')
  expect_warning(listDomains(src.fake, '/shared/bioconductor/'), "bad http request")
 } else TRUE
})

context("HSDSFile")
test_that("Files can be opened for reading", {
 if (check_hsds()) {
  src.hsds <- HSDSSource('https://alserglab.wustl.edu/hsds/')
  f1 <- HSDSFile(src.hsds, '/shared/bioconductor/patelGBMSC.h5')
  expect_equal(f1@type, "domain")
  dsts <- listDatasets(f1)
  expect_true('/assay001' %in% dsts)
  # catch exception: non-existent or empty file
  expect_error(HSDSFile(src.hsds, '/shared/bioconductor/patelGBMSC-nonex.h5'), "Not Found")
  } else TRUE
})

test_that("HSDSFile can be a directory", {
  if (check_hsds()) {
    src.hsds <- HSDSSource('https://alserglab.wustl.edu/hsds/')
    expect_no_warning(f1 <- HSDSFile(src.hsds, '/shared/bioconductor'))
    expect_equal(f1@type, "folder")
    subdomains <- listDomains(f1@src, f1@domain)
    expect_true("/shared/bioconductor/patelGBMSC.h5" %in% subdomains)
  } else TRUE
})

context("HSDSDataset")
test_that("Data can be retrieved from Datasets", {
 if (!check_hsds()) return(TRUE) else {
  src.hsds <- HSDSSource('https://alserglab.wustl.edu/hsds/')
  f2 <- HSDSFile(src.hsds, '/shared/bioconductor/patelGBMSC.h5')
  d2 <- HSDSDataset(f2, '/assay001')
  R <- c(1964617,1266834,1627203,1338108)
  A <- apply(round(getData(d2, c('1:4', '1:65218'), transfermode='JSON')), 1, sum)
  clRA = function(R,A) max(abs(R-A))<1e-6
  expect_true(clRA(R,A))
  A <- apply(round(getData(d2, c('1:4', '1:65218'), transfermode='binary')), 1, sum)
  expect_true(clRA(R,A))
  A <- apply(round(d2[1:4, 1:65218]), 1, sum)
  expect_true(clRA(R,A))
 }
})

context("DelayedArray subclass HSDSArray")
test_that("DelayedArray can be instantiated and accessed",  {
 if (!check_hsds()) return(TRUE) else {
  R <- c(1964617,1266834,1627203,1338108)
  da <- HSDSArray('https://alserglab.wustl.edu/hsds/', 'hsds', 
        '/shared/bioconductor/patelGBMSC.h5', '/assay001')
  A <- apply(round(da[,1:4]),2,sum)
  clRA = function(R,A) max(abs(R-A))<1e-6
  expect_true(clRA(R,A))
 }
})

#context("Four-dimensional datasets")
#test_that("Higher-dimensional dataset access works correctly",  {
# if (!check_hsds()) return(TRUE) else {
#  src <- HSDSSource('http://hsdshdflab.hdfgroup.org')
#  rd <- HSDSDataset(HSDSFile(src, '/home/spollack/testone.h5'), '/group0/group1/group2/data4d')
#  A <- getData(rd, list(3:4, 8:9, 5:6, 2:3))
#  expect_true(sum(A) == 697)
#  dt <- HSDSDataset(HSDSFile(src, '/home/spollack/testone.h5'), '/group0/group1/dataR')
#  B <- getData(dt, list(c(4), c(2, 3, 5, 6), c(5), 1:3))
#  R <- array(c(3140, 3240, 3440, 3540, 3141, 3241, 3441, 3541, 3142, 
#      3242, 3442, 3542), dim=c(4,3))
#  expect_true(all(B == R))
# }
#})

context("Decomposition into slices")
test_that("Bad slices rejected",  {
 if (!check_hsds()) return(TRUE) else {
  tf <- rhdf5client:::checkSlices(c(10, 20, 30), c('5:', ':', ':8'))
  ok <- c('4:10:1', '0:20:1', '0:8:1')
  expect_true(all(unlist(tf) == ok))
  expect_error(rhdf5client:::checkSlices(c(10, 20, 30), c('5:20', ':', ':8'),
    regexp='stop out of range'))
  expect_error(rhdf5client:::checkSlices(c(10, 20, 30), c('10:5', ':', ':8'),
    regexp='slice stop less than slice start'))
  expect_error(rhdf5client:::checkSlices(c(10, 20, 30), c('5:10,0.5', ':', ':8'),
    regexp='malformed slice'))
  }
})


context("String support")
test_that("Basic string support",  {
    src.hsds <- HSDSSource('https://alserglab.wustl.edu/hsds/')
    f <- HSDSFile(src.hsds, "/shared/bioconductor/test_string.h5")
    d <- HSDSDataset(f, "/d")
    expect_true(d@type$class == "H5T_STRING")
    
    v1 <- d[1:10]
    expect_equal(class(v1), "character")
    
    v2 <- d[1]
    expect_equal(class(d[1]), "character")
    expect_equal(v1[1], v2)
})

context("Compound support")
test_that("Basic compound support", {
  src.hsds <- HSDSSource('https://alserglab.wustl.edu/hsds/')
  f <- HSDSFile(src.hsds, "/shared/bioconductor/test_compound.h5")
  d <- HSDSDataset(f, "/d")
  
  expect_equal(d@type$class, "H5T_COMPOUND")
  expect_equal(d@type$fields[[1]]$name, "intCol")
  expect_equal(d@type$fields[[1]]$type$class, "H5T_INTEGER")
  expect_equal(d@type$fields[[2]]$name, "strCol")
  expect_equal(d@type$fields[[2]]$type$class, "H5T_STRING")

  v1 <- d[1]  
  expect_true(is(v1, "data.frame")) # data.table inherits data.frame
  expect_equal(nrow(v1), 1)
  expect_equal(v1$intCol, 1)
  expect_equal(v1$strCol, "a")
  
  v2 <- d[1:2]  
  expect_equal(nrow(v2), 2)
  expect_equal(v2$intCol[1], v1$intCol)
  expect_equal(v2$strCol[1], v1$strCol)
  
  typ <- list(class="H5T_COMPOUND", 
              fields=list(
                list(name="f1", type=list(class="H5T_STRING")),
                list(name="f2", type=list(class="H5T_STRING")),
                list(name="f3", type=list(class="H5T_STRING"))
              ))
  # JSON arrays are used when all columns are strings
  str <- c('[["asd", "qwe", "zxc"], ["a", "b", "c"]]')
  dt <- extractCompoundJSON(type = typ, rjson::fromJSON(str))
  expect_true(is(dt, "data.frame"))
})

context("Scalar support")
test_that("Support of scalar values", {
  src.hsds <- HSDSSource('https://alserglab.wustl.edu/hsds/')
  f <- HSDSFile(src.hsds, "/shared/bioconductor/test_scalar.h5")
  d <- HSDSDataset(f, "/d")
  v <- d[1] 
  expect_true(is(v, "character"))
  expect_identical(v, "I'm scalar")
})

context("Integer type support")
test_that("Integer types are supported properly", {
  src.hsds <- HSDSSource('https://alserglab.wustl.edu/hsds/')
  f <- HSDSFile(src.hsds, "/shared/bioconductor/test_numbers.h5")
  # not checking the actual values!
  
  d <- HSDSDataset(f, "/d_u32")
  v <- d[1:d@shape] 
  expect_equal(length(v), 4)
  expect_equal(v[2], 2**31 - 1)
  
  d <- HSDSDataset(f, "/d_i64")
  v <- d[1:d@shape] 
  expect_equal(length(v), 5)
  expect_equal(v[2], 2**31 - 1)
  
  d <- HSDSDataset(f, "/d_u64")
  v <- d[1:d@shape] 
  expect_equal(length(v), 6)
  expect_equal(v[2], 2**31 - 1)
  
})


test_that("Request errors are reported", {
  skip_if_not(check_hsds())
  
  src.hsds <- HSDSSource("https://developer.nlr.gov/api/hsds")
  expect_error(HSDSFile(src.hsds, "/shared/NASA/NCEP3/ncep3.h5"), "api_key")
})

test_that("Files on a directory on HSDS server can be listed", {
  skip_if_not(check_hsds())
  
  url <- 'https://alserglab.wustl.edu/hsds/?domain=/counts'
  expect_true(!is.null(getHSDSFileList(url)))
})

if (FALSE) {
context("String support")
test_that("Basic string support",  {
    src.hsds <- HSDSSource('http://hsdshdflab.hdfgroup.org')
    f <- HSDSFile(src.hsds, "/shared/NREL/sample/windspeed_z5.h5")
    d <- HSDSDataset(f, "/datetime")
    expect_true(d@type$class == "H5T_STRING")
    
    v1 <- d[1:10]
    expect_equal(class(v1), "character")
    
    v2 <- d[1]
    expect_equal(class(d[1]), "character")
    expect_equal(v1[1], v2)
})

context("Compound support")
test_that("Basic compound support", {
  src.hsds <- HSDSSource('http://hsdshdflab.hdfgroup.org')
  f <- HSDSFile(src.hsds, "/shared/ghcn/year/inventory.h5")
  d <- HSDSDataset(f, "/inventory")
  
  expect_equal(d@type$class, "H5T_COMPOUND")
  expect_equal(d@type$fields[[1]]$name, "year")
  expect_equal(d@type$fields[[1]]$type$class, "H5T_STRING")

  v1 <- d[1]  
  expect_true(is(v1, "data.frame")) # data.table inherits data.frame
  expect_equal(nrow(v1), 1)
  expect_equal(v1$year, "1763")
  expect_equal(v1$loadstart, 1636742475)
  
  v2 <- d[1:2]  
  expect_equal(nrow(v2), 2)
  expect_equal(v2$year[1], v1$year)
  expect_equal(v2$loadstart[1], v1$loadstart)
})

context("Scalar support")
test_that("Support of scalar values", {
  src.hsds <- HSDSSource('http://hsdshdflab.hdfgroup.org')
  f <- HSDSFile(src.hsds, "/shared/NASA/NCEP3/ncep3.h5")
  d <- HSDSDataset(f, "/HDFEOS INFORMATION/StructMetadata.0")
  v <- d[1] 
  expect_true(is(v, "character"))
  expect_true(startsWith(v, "GROUP"))
})
}

library(rhdf5)

h5file <- "test_string.h5"
h5createFile(h5file)
d <- as.character(1:10)
h5write(d, h5file, "d")
h5ls(h5file)

h5file <- "test_compound.h5"
h5createFile(h5file)
d <- data.frame(intCol=1:3, strCol=c("a", "b", "c"))
h5write(d, h5file, "d")
h5ls(h5file)

h5file <- "test_scalar.h5"
h5createFile(h5file)
d <- "I'm scalar"
h5write(d, h5file, "d")
h5ls(h5file)

#' HSDSArraySeed for HSDSArray backend to DelayedArray
#'
#' @name HSDSArraySeed
#' @slot endpoint URL of remote server
#' @slot svrtype type of server, must be either 'hsds' or 'h5serv'
#' @slot domain HDF5 domain of H5 file on server
#' @slot dsetname complete internal path to dataset in H5 file
#' @slot dataset object of type HSDSDataset for access to the H5 dataset
#' @aliases HSDSArraySeed-class
#' @import DelayedArray
# #' @exportClass HSDSArraySeed
setClass("HSDSArraySeed", contains=c("Array"), 
  slots = c(endpoint="character",    # URL
            svrtype="character",     # 'h5serv' or 'hsds'
            domain="character",      # extra-file (file system) path
            dsetname="character",    # complete intra-file path
            dataset="HSDSDataset")       # HSDSDataset object 
)

#' Construct an object of type HSDSArraySeed 
#' 
#' @name HSDSArraySeed
#' 
#' @param endpoint URL of remote server
#' @param svrtype type of server, must be either 'hsds' or 'h5serv'
#' @param domain HDF5 domain of H5 file on server
#' @param dsetname complete internal path to dataset in H5 file
#' @return An initialized object of type HSDSArraySeed
HSDSArraySeed <- function(endpoint, svrtype, domain, dsetname)  {
  src <- .HSDSSource(endpoint, svrtype)  # IF YOU ARE DEPRECATING IT YOU SHOULD NOT USE IT
  fle <- .HSDSFile(src, domain)  # private
  dset <- .HSDSDataset(fle, dsetname)
  new("HSDSArraySeed", endpoint=endpoint, svrtype=svrtype, domain=domain, 
      dsetname=dsetname, dataset=dset)
}

#' Obtain names of dimensions for an object of type HSDSArraySeed
#' 
#' (required by DelayedArray seed contract, returns NULL list)
#' 
#' @name dimnames
#' @param x An object of type HSDSArraySeed
#' @return A NULL list of length equal to the array dimensionality
#' @aliases dimnames,HSDSArraySeed-method
setMethod("dimnames", "HSDSArraySeed", function(x)  {
  n <- length(x@dataset@shape)
  rt <- vector(mode="list", length=n)   # null list
})

#' Obtain dimensions of an object of type HSDSArraySeed
#' 
#' (required by DelayedArray seed contract)
#' HDF server content is assumed transposed relative to R matrix layout.
#' This anticipates H5 datasets on the server with rows for 
#' experimental samples and columns for *-omic features. The 
#' Bioconductor SummarizedExperiment requires *-omic features in 
#' rows and samples in columns.
#' 
#' @name dim
#' @param x An object of type HSDSArraySeed
#' @return A numeric vector of the dimensions
#' @aliases dim,HSDSArraySeed-method
setMethod("dim", "HSDSArraySeed", function(x)  {
   as.integer(rev(x@dataset@shape))   
})

#' Access dataset backed by an HSDSArraySeed
#'
#' @name extract_array
#' @param x An object of type HSDSArraySeed
#' @param index A list of numeric vectors to be accessed, one vector 
#' for each dimension of the array object. A NULL vector indicates
#' the entire range of indices in that dimension. A zero-length
#' vector indicates no indices in the relevant dimension. (Accordingly,
#' any zero-length vector of indices will result in an empty array
#' being returned.)
#' 
#' @return An array containing the data elements corresponding to the 
#' indices requested
#' @aliases extract_array,HSDSArraySeed-method
#' 
# #' @export
#'
# TODO: seed contract requires repeated and descending indices 
# (e.g., c(1,2,3,3,2,1)) be returned correctly. 
setMethod("extract_array", "HSDSArraySeed", function(x, index)  {
  index <- rev(index)

  # two special cases
  # (i) NULL index - signifies all elements in this dimension
  # (ii) zero-length index - signifies zero elements in this dimension
  # which means a null fetch 

  idxlist <- lapply(seq_along(index), 
    function(i)  {
      if (is.null(index[[i]]))  {
        n <- x@dataset@shape[i]
        if (n == 0)  {
          v <- numeric(0)
        } else  {
          v <- seq(1,n)
        }
      } else if (length(index[[i]]) == 0)  {
        v <- numeric(0)
      } else  {
        v <- unlist(slicelst(index[[i]]))
      }
      v
    }) 

  rdims <- lapply(idxlist, function(v) length(v))
  nullfetch <- any(rdims == 0)
  if (nullfetch) { 
    A <- array(numeric(0), dim=rdims)
  } else  {
    A <- getDataList(x@dataset, idxlist)
    # unflatten vector if necessary: see note at 
    # end of Dataset::getDataVec on flattened result
    # for fetch of single-width dimensioned arrays.
    A <- array(A, dim=rdims)
  }
  R <- t(A)   # untranspose the transpose
  R
})

#' A DelayedArray backend for accessing a remote HDF5 server.
#' 
#' @name HSDSArray 
#' @family  HSDSArray
#' @aliases HSDSArray-class
#' @exportClass HSDSArray
setClass("HSDSArray", contains="DelayedArray")

#' DelayedMatrix subclass for a two-dimensional HSDSArray
#'
#' @name HSDSMatrix
#' @family  HSDSArray
#' @aliases HSDSMatrix-class
#' @exportClass HSDSMatrix
setClass("HSDSMatrix", contains=c("DelayedMatrix", "HSDSArray"))

setMethod("matrixClass", "HSDSArray", function(x) "HSDSMatrix")

setAs("HSDSArray", "HSDSMatrix", function(from)  { 
    new("HSDSMatrix", from)
  }
)

#' Coercion method from HSDSMatrix to its superclass HSDSArray
#'
#' @name as
#' @family  HSDSArray

setAs("HSDSMatrix", "HSDSArray", function(from) from)   # no-op

#' @importFrom DelayedArray new_DelayedArray
setMethod("DelayedArray", "HSDSArraySeed", 
  function(seed)   
    new_DelayedArray(seed, Class="HSDSArray") 
)

#' a test request
#' @return logical, TRUE if hsds behaving as expected
#' @examples
#' check_hsds()
#' @export
check_hsds <- function() {
 tst <- "https://alserglab.wustl.edu/hsds/?domain=/shared/bioconductor/patelGBMSC.h5"
 ans <- httr::GET(tst)
 return(ans$status_code == 200)
}

#' Construct an object of type HSDSArray directly from the data
#' members of its seed
#'
#' @param endpoint URL of remote server
#' @param svrtype type of server, must be either 'hsds' or 'h5serv'
#' @param domain HDF5 domain of H5 file on server
#' @param dsetname complete internal path to dataset in H5 file
#' @return An initialized object of type HSDSArray
#' @examples
#' if (check_hsds()) {
#'  HSDSArray(URL_hsds(), 
#'     "hsds", "/shared/bioconductor/patelGBMSC.h5", "/assay001")
#' }
#' @export
HSDSArray <- function(endpoint, svrtype, domain, dsetname)  {
  DelayedArray(HSDSArraySeed(endpoint, svrtype, domain, dsetname))
}

deprecate_msg = paste0("This function is deprecated. The new interface to rhdf5client",
 " is exclusively through its DelayedArray backend HSDSArray")

# utilities for index processing
#'
#' This function is deprecated and will be defunct in the next release.
#'
# sproc(isplit(vec)) will convert vec representing R integer vector
# into a list of HDF5server 'select' index candidates
#' isplit converts a numeric vector into a list of sequences for compact reexpression
#' @name isplit
#' @rdname sproc
#' @import methods
#' @param x a numeric vector (should be integers)
#' @return list of vectors of integers which can be expressed as initial/final/stride triplets 
#' @export
isplit = function(x)  {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  if (length(x)==1) return(list(`1`=x))
  y <- rep(0,length(x))
  i <- 3
  while (i <= length(x))  {
    if (x[i]-x[i-1] == x[i-1]-x[i-2])  { 
      i <- i + 1
    } else  {
      y[i] <- 1
      i <- i + 2
    }
  }
  grps <- cumsum(y)+1
  split(x, grps)
}

nosci <- function(x) format(x, scientific=FALSE)

#' sproc makes vector of type character of triplets initial:final:stride in R-conventions
#' @name sproc
#' @rdname sproc
#' @param spl output of isplit
#' @return list of colon-delimited strings each with initial/final/stride triplet 
#' @examples
#' inds = c(1:10, seq(25,50,2), seq(200,150,-2))
#' sproc(isplit(inds))
#' @export
sproc = function(spl)  {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  ans <- lapply(spl, function(x) {
    if (length(x) == 1) return(paste(nosci(x), ":", nosci(x), ":1", sep=""))
    d <- x[2]-x[1]
    paste(nosci(x[1]),":",nosci(x[length(x)]),":",d, sep="")
  })
  ans <- r2py(ans)
  ans
}

# r2py makes R-convention strings into python-convention strings
# HDF5 server uses python conventions
# R conventions: array indices begin at 1, and ranges include last element
# python conventions: array indices begin at 0 and ranges exclude last element
# 
# HDF5 server does not allow decreasing index ranges. Consistency requires 
# the range c(7, 6, 5, 4, 3) should have the same elements as c(3, 4, 5, 6, 7)
# except flipped. r2py maps "3:7:1" to "2:7:1", so it maps "7:3:-1" to 
# "7:2:-1". rhdf5client will flip this to "3:7:1", fetch the data, then flip it back. 
r2py <- function(spr)  {
  ans <- lapply(spr, function(x)  {
    v <- as.numeric(unlist(strsplit(x, ":")))
    if (v[3] > 0)  {
      paste(nosci(v[1]-1), ":", nosci(v[2]), ":", nosci(v[3]), sep="")
    } else  {
      paste(nosci(v[1]), ":", nosci(v[2]-1), ":", nosci(v[3]), sep="")
    }
  })
  ans 
}

# ivindx reverses a vector of negative indices into a vector of positive indices
# example: ii <- ivindx(-c(2, 4, 8), 10)    # c(1, 3, 5, 6, 7, 8, 9)
ivindx <- function(iidx, didx)  {
  ii <- -iidx
  v <- 1:didx
  v[ !(v %in% ii) ]
}

#' rhdf5client: A package for accessing HDFGroup HDF5 servers from R.
#'
#' The rhdf5client package provides read-only access to HDF5 files
#' maintained on a server. The HDFGroup provides two servers, an
#' obsolescent one called `h5serv` and the newer prototype called `hsds`.  
#' @docType package
#' @name rhdf5client
NULL

#' @import methods

# private to deal with distinct behavior of httr::GET on windows
winpref = function(x, pref="http://") {
  if (length(grep("^http:", x))==0 & length(grep("^https:", x))==0) x = paste0(pref,x)
  x
}

deprecate_msg = paste0("This function is deprecated. The new interface to rhdf5client",
 " is exclusively through its DelayedArray backend HSDSArray")

#' An S4 class to represent a HDF5 server listening on a port.
#'
#' This class is deprecated and will be defunct in the next release.
#'
#' @slot endpoint URL for server 
#' @slot type Type of server software at the source; must be 
#' either 'h5serv' or (default) 'hsds' 
setClass("HSDSSource", representation(endpoint="character", type="character"))

#' Construct an object of type HSDSSource.
#'
#' A HSDSSource is a representation of a URL which provides access to a HDF5 
#' server (either h5serv or hsds.) 
#'
#' This function is deprecated and will be defunct in the next release.
#'
#' @name HSDSSource
#' @param endpoint URL for server 
#' @param type Type of server software at the source; must be 
#' @return An object of type HSDSSource
#' @examples
#' if (check_hsds()) {
#'  src.hsds <- HSDSSource(URL_hsds())
#' }
#' @export
HSDSSource <- function(endpoint, type='hsds')  {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  if (!(type %in% c('h5serv', 'hsds')))
    stop(paste("unknown server type ", type))
  obj <- new("HSDSSource", endpoint=endpoint, type=type)
  # member root id also?
}


.HSDSSource <- function(endpoint, type='hsds')  { # after deprecation cycle this private function is used
  if (!(type %in% c('h5serv', 'hsds')))
    stop(paste("unknown server type ", type))
  # obj <- new("HSDSSource", endpoint=endpoint, type=type) # should not use assignment as function value
  new("HSDSSource", endpoint=endpoint, type=type)
}

#' List files and subdirectories of a domain
#'
#' The user needs to give the domain to start in. The search
#' will be non-recursive. I.e., output for domain '/home/jreadey/' will 
#' not return the files in '/home/jreadey/HDFLabTutorial/'
#'
#' This function is deprecated and will be defunct in the next release.
#'
#' @param object An object of type HSDSSource 
#'
#' @param rootdir A slash-separated directory in the HSDSSource file system. 
#'
#' @return a vector of domains in the rootdir
#'
#' @export
#' @docType methods
#' @rdname listDomains-methods
#'
#' @examples
#' src.hsds <- HSDSSource(URL_hsds())
#' listDomains(src.hsds, '/shared')
setGeneric('listDomains', function(object, rootdir) standardGeneric('listDomains'))

#' @rdname listDomains-methods
#' @aliases listDomains,HSDSSource,character-method
setMethod('listDomains', c("HSDSSource", "character"), 
  function(object, rootdir)  {
    #.Deprecated("HSDSArray", NULL, deprecate_msg)
    ll <- domainContents(object, rootdir)
    vapply(ll, function(l) l$filename, character(1))
  })

#' @rdname listDomains-methods
#' @aliases listDomains,HSDSSource,missing-method
setMethod('listDomains', c("HSDSSource", "missing"),  
  function(object) { 
    #.Deprecated("HSDSArray", NULL, deprecate_msg)
    listDomains(object, '/hdfgroup/org') 
  })

# private
domainContents <- function(object, rootdir = '/hdfgroup/org')  {
  if (!(is(object, "HSDSSource")))
    stop("getDomains called on a non-HSDSSource object")

  # result list
  rlistsz <- 1000
  rlist <- vector("list", rlistsz)
  nrlist <- 1

  append_to_results <- function(fn, ft)   {
    rlist[[nrlist]] <<- list(filename=fn, filetype=ft)
    nrlist <<- nrlist + 1
    if (nrlist >= rlistsz)  
      stop("list overflow")
  }

#  The h5serv API does not have a GET /domains method, so we need
#  to descend the tree to the requested domain node by UUID's.

  result <- tryCatch({
    if (object@type == 'h5serv')  {
      append_if_h5_file <- function(ll, ff)  {
        if ('h5domain' %in% names(ll)) {
          ff <- c(ff, ll[['h5domain']])
        } 
        ff
      }

      # dissect the path. assume the last two elements are the root domain. 
      s <- .Platform$file.sep
      if (substr(rootdir, 1, 1) != s) rootdir <- paste0(s, rootdir)
      pth <- strsplit(rootdir, s)[[1]]
      pth <- pth[-1]      # empty string before leading '/'
      n <- length(pth)
      dstr <- paste0('.', pth[n-1], '.', pth[n])   # root domain
      if (length(pth) == 2)
        pth = c()
      else
        pth <- pth[1:(n-2)]

      request <- paste0(object@endpoint, "/")
      response <- submitRequest(request)
      nextid <- response[['root']]
      link <- list()

      while (length(pth) > 0)  {

        request <- paste0(object@endpoint, "/groups/", nextid, "/links")
        response <- submitRequest(request)
        links <- response[['links']]

        v <- vapply(links, function(lk) { 
          'title' %in% names(lk) && lk[['title']] == pth[length(pth)] }, 
          logical(1))
        if (any(v))  {
          link <- links[[which(v)]]
          nextid <- link[['id']]
          dstr <- paste0(pth[length(pth)], '.', dstr)
          pth <- pth[-length(pth)]
        }
        else  {    # special case: rootdir is a file 
          v <- vapply(links, function(lk) { 
            'class' %in% names(lk) && lk[['class']] == 'H5L_TYPE_EXTERNAL' && 
            'h5domain' %in% names(lk) && lk[['h5domain']] == pth[length(pth)] }, 
            logical(1))
          if (any(v))  {
            link <- links[[which(v)]]
            pth <- c()
          } else  {
            stop(paste0("domain ",rootdir, " not found"))
          }
        }
      }
  
      if (length(link) == 0 || 
          link[['class']] == 'H5L_TYPE_HARD')  {  # domain is a directory
        request <- paste0(object@endpoint, "/groups/", nextid, "/links")
        response <- submitRequest(request)
        links <- response[['links']]
        for (lk in links)  {
          if (lk[['class']] == 'H5L_TYPE_HARD')  {             # a subdirectory
            append_to_results(paste0(lk[['title']], rootdir), 'directory')
          } else if (lk[['class']] == 'H5L_TYPE_EXTERNAL')  {  # a file
            append_to_results(lk[['h5domain']], 'file')
          }
        }

      } else  {  # 'self' signals that rootdir *is* a file 
        append_to_results(rootdir, 'self')
      }
      1
    }
    else  { # 'hsds'

      # no postpended slash gets the type of the object
      request <- paste0(object@endpoint, "/domains?domain=", rootdir)
      response <- submitRequest(request)
      if ('domains' %in% names(response) && length(response[['domains']]) > 0)  {

        # with postpended slash gets directory contents
        # if ('root' %in% names(response[['domains']][[1]]) && 

        domains <- response[['domains']]
        for (domain in domains)  {
          if ('class' %in% names(domain) && 'name' %in% names(domain) && 
             (domain[['class']] == 'domain' || domain[['class']] == 'folder'))  {
            fn <- domain[['name']]
            ft <- 'directory'
            if (domain[['class']] == 'domain') 
              ft <- 'file'
            append_to_results(fn, ft)
          }
        }
      }
      1
    }
  }, error=function(e) { -1 })    # catch http request errors
  if (result == -1)  {
    warning("bad http request", call. = FALSE)
  }

  rlist[-which(sapply(rlist, is.null))]

}

# private - submit request and handle errors 
submitRequest <- function(req, transfermode='JSON')  {
  rsp <- tryCatch({
      if (transfermode == 'JSON')  {
        if (.Platform$OS.type == "windows") req = winpref(req)
        httr::GET(req)
      } else if (transfermode == 'binary')  {
        if (.Platform$OS.type == "windows") req = winpref(req)
        httr::GET(req, add_headers(Accept="application/octet-stream"))
      }
    }, error=function(e) { NULL }
  )

  if (is.null(rsp))  
    stop("Bad request (no response)")

  if ('status_code' %in% names(rsp) && rsp$status_code != 200)  
    stop("Request failed:\nrequest: ", req, "\n",
    "error message: ", readBin(rsp$content, what="character"))

  if (transfermode == 'JSON')  {
    rsp <- tryCatch({
      rjson::fromJSON(readBin(rsp$content, what="character"))
    }, error=function(e) { NULL })
  }

  if (is.null(rsp))  
    stop("Bad request")

  # Note: rsp could be the empty string - not an error
  # maybe throw a warning condition?

  return(rsp)

}

#' manage hsds URL
#'
#'
#' @return URL of hsds server
#' @examples
#' URL_hsds()
#' @export
URL_hsds = function() {
  "http://hsdshdflab.hdfgroup.org"
}


#' H5S_source identifies an HDF5/HSDS server and manages some metadata about contents
#'
#' This class is deprecated and will be defunct in the next release.
#'
#' @name H5S_source
#' @rdname H5S_source-class
#' @slot serverURL character string with a URL
#' @slot dsmeta DataFrame instance with metadata about content of h5serv server
#' @slot dmains DataFrame instance with metadata about the content of hsds server
#' @slot getReq DataFrame instance with metadata about hsds server
#' @slot folderPath character string with path to user's folder/file on hsds server
#' @aliases H5S_source-class
#' @exportClass H5S_source

setClass("H5S_source", representation(serverURL="character", dsmeta="DataFrame", getReq = "DataFrame",dmains= "DataFrame", folderPath = "character"))
setMethod("show", "H5S_source", function(object) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  if(nrow(object@dsmeta)>1){
  cat(" H5serv server url : ", object@serverURL, 
      "\n There are", nrow(object@dsmeta), "groups.",
      "\n Use groups(), links(), ..., to probe and access metadata.",
      "\n Use dsmeta() to get information on datasets within groups.",
      "\n Use [[ [dsname] ]]  to get a reference suitable for [i, j] subsetting.\n")
  }
  else{
  cat(" HSDS server url :", object@serverURL,
      "\n Use getReq() to get information on the server.",
      "\n Use setPath() to specify path to a hdf5 file.",
      "\n Use fetchDatasets() to get id of the dataset of interest.\n")
    }
})

# private: extract hostname from a file URL (return null for group URL)
fixtarget = function(x) sub(".*host=(.*).h5s.channingremotedata.org", "\\1", x)

#' construct H5S_source
#'
#' This function is deprecated and will be defunct in the next release.
#'
#' @name H5S_source
#' @rdname H5S_source-class
#' @param serverURL a URL for a port for HDF5Server
#' @param domain character string with path to file for HSDS 
#' @param \dots not used
#' @note The dsmeta slot holds a DataFrame with a column \code{dsnames}
#' that is a list with ith element a character vector of all dsnames
#' available for the ith group.  There is no effort at present to
#' search all groups for candidate datasets.
#' @note If the domain for the HSDS server is known, 
#' pass the domain path as a character string along with ther serverURL
#' @return an initialized object of type H5S_source
#' @examples
#' \dontrun{
#' bigec2 = H5S_source(URL_h5serv()) # h5serv 
#' bigec2
#' dsmeta(bigec2)[1:2,]       # two groups
#' dsmeta(bigec2)[1,2][[1]]   # all dataset candidates in group 1
#' hsdsCon = H5S_source(URL_hsds()) # hsds server connection
#' hsdsCon
#' getReq(hsdsCon)
#' setPath(hsdsCon,"/home/stvjc/hdf5_mat.h5") -> hsds
#' fetchDatasets(hsds)     # grab the dataset id of interest 
#' H5S_dataset2(hsds, "d-a9e4b71c-8ea2-11e8-9306-0242ac120022")
#' }
#' @export
H5S_source = function(serverURL, domain, ...) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  if(missing(domain)){
    serverCheck = serverVersion(serverURL)
    if(serverCheck == 1){
      tmp <- new("H5S_source", serverURL=serverURL, dsmeta=DataFrame())
      grps <- groups(tmp)
      message("analyzing groups for their links...")
      thel <- targs <- List(targs=lapply( seq_len(nrow(grps)), 
                                          function(x) fixtarget(hosts(links(tmp,x)))))
      message("done")
      dsm <- DataFrame(groupnum=seq_len(nrow(grps)), dsnames=thel, grp.uuid=grps$groups)
      obj <- new("H5S_source", serverURL=serverURL, dsmeta=dsm)
      obj
    }
    else{
      tmp <- new("H5S_source", serverURL=serverURL, getReq=DataFrame())
      get <- hsdsInfo(tmp)
      get.df <- DataFrame(get)
      obj <- new("H5S_source", serverURL=serverURL, getReq=get.df )
    }
  }
  else{
    tmp <- new("H5S_source", serverURL=serverURL, getReq=DataFrame())
    get <- hsdsInfo(tmp)
    get.df <- DataFrame(get)
    obj <- new("H5S_source",serverURL=serverURL, getReq=get.df, folderPath=domain)
    #H5S_dataset2(obj)
  }
}

#' list information about datasets available in an H5S_source
#'
#' This function is deprecated and will be defunct in the next release.
#'
#' @param src H5S_source instance
#' @return data frame with one row for each group and three columns. The 
#' second column has the list of datasets in the group.
#' @examples
#' \dontrun{
#' bigec2 = H5S_source(URL_h5serv())
#' dsm <- dsmeta(bigec2) 
#' dst <- unlist(dsm[1,2])    # all dataset candidates in group 1
#' }
#' @export
dsmeta = function(src) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  src@dsmeta
}

#' list information about server content available in an H5S_source hsds instance
#'
#' This function is deprecated and will be defunct in the next release.
#'
#' @param src H5S_source instance
#' @return data frame with 5 columns for one row for each user's data
#' @export
getReq = function(src) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  src@getReq
}

#' Subscript operator
#'
#' This function is deprecated and will be defunct in the next release.
#'
#' @name [[
#' @aliases [[,H5S_source,character-method
#' @param x instance of H5S_source
#' @param i character string intended to identify dataset on server
#' @param j not used
#' @exportMethod '[['
setMethod("[[", c("H5S_source", "character", "ANY"), function(x, i, j) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  dataset(x, i)
})

#' HDF5 server data groups accessor
#' @exportMethod groups
#' @export groups
#' @docType methods
#' @rdname groups-methods
setGeneric("groups", function(object, index, ...) standardGeneric("groups"))

#' @param object H5S_source instance
#'
#' This function is deprecated and will be defunct in the next release.
#'
#' @param index numeric, if present, extracts metadata about selected group (sequential ordering 
#' of groups as returned by server) access for group information for HDF5 server
#' @param \dots not used
#' @return a data frame with group name and number of links for each group
#' @examples
#' \dontrun{
#' bigec2 = H5S_source(URL_h5serv())
#' groups(bigec2)
#' }
#' @rdname groups-methods
#' @aliases groups,H5S_source,missing-method 
setMethod("groups", c("H5S_source", "missing"), function(object, index, ...) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  target = paste0(.serverURL(object), "/groups")
  ans = transl(target) 

  # find the root group: hh is a matrix with first column "href" and second "rel"
  # Find the one whose "rel" column is "root" and pull the name out of the "href"

  if (length(ans$hrefs)<1)  { 
    stop("no group table at server URL")
  }
  hh = t(sapply(ans$hrefs, force))     
  rootgroup = sub(".*groups.", "", hh[hh[,2]=="root", 1])  
  grps = c(rootgroup, unlist(ans$groups))   # all groups plus the root group

  # get the link count for each group
  nl = vapply(1:length(grps), function(x) {
    gname = grps[x]
    target = paste0(.serverURL(object), "/groups/", gname, "/links" )
    ans = transl(target)  
    length(ans$links)
  }, integer(1))
  DataFrame(groups=grps, nlinks=nl)
})

#' @aliases groups,H5S_source,numeric-method
#' @rdname groups-methods
setMethod("groups", c("H5S_source", "numeric"), function(object, index, ...) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  groups(object)[index,,drop=FALSE]
})

setClass("H5S_linkset", representation(links="list", group="character",
                                       source="H5S_source"))
setMethod("show", "H5S_linkset", function(object) {
  cat("HDF5 server link set for group", object@group, "\n")
  cat(" There are", length(object@links$links), "links.\n")
  cat(" Use targets([linkset]) to extract target URLs.\n")
})

#' set path for hsds server resource
#' 
#'  This function is deprecated and will be defunct in the next release.
#' 
#' @param object H5S_source instance
#' @param folderPath character string with path to user's folder on hsds server
#' @param \dots not used
#' @return an updated object with folderPath set
#' @examples
#' \dontrun{
#' hsdsCon = H5S_source(URL_hsds()) # hsds server connection
#' setPath(hsdsCon, "/home/stvjc/hdf5_mat.h5")-> hsds
#' }
#' @docType methods
#' @rdname setPath
#' @aliases setPath,H5S_source,character-method
#' @export setPath
#' @exportMethod setPath
setGeneric("setPath", function(object,folderPath, ...) standardGeneric("setPath"))
setMethod("setPath", c("H5S_source","character"), function(object, folderPath, ...) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  object@folderPath = folderPath
  object
})

#' access for link metadata for HDF5 server groups
#'
#' This function is deprecated and will be defunct in the next release.
#'
#' @param object H5S_source instance
#' @param index numeric group index
#' @param \dots not used
#' @return an object of type H5S_linkset with the linkset of the group
#' @examples
#' \dontrun{
#' bigec2 = H5S_source(URL_h5serv())
#' lks <- links(bigec2, 1)    # linkset for root group 
#' urls <- targets(lks)       # URLs of datasets in linkset
#' }
#' @rdname links
#' @aliases links,H5S_source,numeric-method
#' @export links
#' @exportMethod links
setGeneric("links", function(object, index, ...) standardGeneric("links"))
setMethod("links", c("H5S_source", "numeric"), function(object, index, ...) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  gname = groups(object, index)[["groups"]][1] # skirt mcols bug
  target = paste0(.serverURL(object), "/groups/", gname, "/links" )
  ans = transl(target) 
  new("H5S_linkset", links=ans, source=object, group=gname)
})

.links <- function(linkset) { linkset@links }    # private accessor

#' provide the full URLs for link members
#'
#' This function is deprecated and will be defunct in the next release.
#'
#' @param h5linkset instance of H5S_linkset
#' @param index numeric index into link vector - ignored
#' @return a vector of dataset tags
#' @rdname targets
#' @examples
#' \dontrun{
#' bigec2 = H5S_source(URL_h5serv())
#' lks <- links(bigec2, 1)    # linkset for root group 
#' urls <- targets(lks)       # URLs of datasets in linkset
#' }
#' @export

targets = function(h5linkset, index) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  lks <- .links(h5linkset)
  vapply(h5linkset@links$links, function(lk) lk[["target"]], character(1)) 
}

# private: return all URLs in the linkset that link to datasets
# cleanIP: if TRUE, remove the URL up to the host 
# note: index is ignored in targets(), so ignored in hosts()
hosts = function(h5linkset, index, cleanIP=TRUE) {
  ans = targets(h5linkset, index)      
  ans = ans[grep("host=", ans)]
  if (cleanIP) {
    gsub(".*host=", "host=", ans)
  } else {
    ans
  }
}

#' construct H5S_dataset object
#'
#' This function is deprecated and will be defunct in the next release.
#'
#' @name H5S_dataset
#' @import S4Vectors
#' @slot source instance of H5S_source instance
#' @slot simpleName character string naming dataset 
#' @slot shapes list including dimension information
#' @slot hrefs DataFrame of hrefs as defined in the API
#' @slot allatts list of all attributes
#' @slot presel string prepared for select operation in GET
#' @slot transfermode default "JSON" or "binary" for binary transfer
#' @rdname H5S_dataset
#' @aliases H5S_dataset-class
#' @exportClass H5S_dataset
setClass("H5S_dataset", representation(
  source="H5S_source", simpleName="character",
  shapes="list", hrefs="DataFrame", allatts="list", 
  presel="character", transfermode="character"))
setMethod("show", "H5S_dataset", function(object) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  cat("H5S_dataset instance:\n")
  curdim = object@shapes$dims
  print(data.frame(dsname=object@simpleName, intl.dim1=curdim[1], intl.dim2=curdim[2], 
                   created=object@allatts$created, type.base=object@allatts$type$base))
  #cat("Use [[", object@simplename, "]] to acquire reference amenable to [i,j] subsetting.\n")
})

#' replace transfer mode
#'
#' This function is deprecated and will be defunct in the next release.
#'
#' @name transfermode<-
#' @param object instance of H5S_linkset
#' @param value either "JSON" (default) or "binary"
#' @return updated object of type H5S_dataset
#' @rdname transfermode
#' @aliases transfermode<-,H5S_dataset-method
#' @docType methods
#' @export
setGeneric("transfermode<-", def = function(object, value) { standardGeneric("transfermode<-") })
setReplaceMethod("transfermode", "H5S_dataset", 
  function(object, value) {  
    #.Deprecated("HSDSArray", NULL, deprecate_msg)
    if ( value == "JSON" | value == "binary" )  {
      object@transfermode <- value  
    }  else  {
      warning(paste("ignoring request for unknown transfer mode ", value))
    }
    object
  }
)

# private to deal with distinct behavior of httr::GET on windows
winpref = function(x, pref="http://") {
  if (length(grep("^http:", x))==0 & length(grep("^https:", x))==0) x = paste0(pref,x)
  x
}

# private: get content from host
# param targ is the URL with query
transl = function(targ)  {  
  if (.Platform$OS.type == "windows") targ = winpref(targ)
  rsp <- GET(targ)
  if (rsp$status_code != 200)  
    stop(paste("error: can't read JSON ", targ, sep=""))
  jsn <- readBin(rsp$content, what="character")
  fromJSON(jsn)
}

# private: get numeric content from host by binary transfer
# param targ is the URL with query
# param nele is the number of numeric elements expected
bintransl = function(targ, nele, h5typ)  {  
  if (.Platform$OS.type == "windows") targ = winpref(targ)
  rsp <- GET(targ, add_headers(Accept="application/octet-stream"))
  if (rsp$status_code != 200)  
    stop(paste("error: can't read binary ", targ, sep=""))

  if (h5typ == "H5T_IEEE_I32LE")  {
    readBin(rsp$content, what="integer", n = nele, size = NA_integer_, 
      signed = TRUE, endian = "little")
  } else if (h5typ == "H5T_IEEE_F64LE")  {
    readBin(rsp$content, what="double", n = nele, size = 8, 
      signed = TRUE, endian = "little")
  } else  {
    stop(paste0("error: unrecognized data type in binary transfer: ", h5typ))
  }
}

# private: get numeric content from host by JSON transfer
# param targ is the URL with query
# param nele is the number of numeric elements expected (ignored)
jsontransl = function(uu, nele)  {  
  val <- transl(uu)$value
  if (is.list(val))  {
    result <- do.call(c, val)
  } else {
    result <- val
  }
}


#private : to check if the server being called is hsds or h5serv
serverVersion <- function(serverURL = serverURL){
  #validURL = url.exists(serverURL)
  if (.Platform$OS.type == "windows") serverURL = winpref(serverURL)
  if(httr::http_status(GET(serverURL))$reason == "OK"){
    serverResponse = fromJSON(file=serverURL)
    if("root" %in% attributes(serverResponse)$names){
      flag = 1   ### This is a h5serv
      return(flag)
    }
    else{ 
      flag = 2   ### This is a hsds server
      return(flag)
    }
  }
  else{
    flag = 2
    return(flag) # this is the new hsds server
  }
  
}

#' extract elements from H5S_dataset
#'
#' This function is deprecated and will be defunct in the next release.
#'
#' @param x instance of H5S_dataset
#' @param i select option for first matrix index in HDF5 server value API
#' @param j select option for second matrix index in HDF5 server value API
#' @param \dots unused
#' @param drop logical defaults to FALSE
#' @return matrix of data obtained
#' @exportMethod [
setMethod("[", c("H5S_dataset", "numeric", "numeric", "ANY"), function(x, i, j, ..., drop=FALSE) {
# bracket selection passed directly to HDF5 server ... row-major
#
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  ii <- as.integer(i)
  jj <- as.integer(j)
  if (any(ii != i) | any(jj != j))  {
    stop("index is non-integral")
  }

  if (all(i < 0)) {
    i <- ivindx(i, internalDim(x)[1]) 
  }
  if (all(j < 0)) {
    j <- ivindx(j, internalDim(x)[2]) 
  }
  if (any(i <= 0) | any(j <= 0) )  {
    stop("index is negative")
  }
  if (any(i > internalDim(x)[1]) | any(j > internalDim(x)[2]))  {
    stop("index is negative")
  }

  ind1 <- sproc(isplit(i))
  ind2 <- sproc(isplit(j))

  if (length(ind1) == 1 & length(ind2) == 1)  {
    ans <- t(x[ ind1[[1]], ind2[[1]] ])
  } else if (length(ind2) == 1)  {
    ansl <- lapply(ind1, function(i1) t(x[i1, ind2[[1]]]))
    ans <- do.call(cbind, ansl)
  } else if (length(ind1) == 1)  {
    ansl <- lapply(ind2, function(i2) t(x[ind1[[1]], i2]))
    ans <- do.call(rbind, ansl)
  } else  {
    ansl <- lapply(ind1, function(i1)  {
      do.call(rbind, lapply(ind2, function(i2)  {
        t(x[i1, i2])
      }))
    })
    ans <- do.call(cbind, ansl)
  }
  t(ans)
})
  

#' extract elements from H5S_dataset
#' @aliases [,H5S_dataset,character,character-method
#' @param x instance of H5S_dataset
#' @param i character vector of row selections
#' @param j character vector of column selections
#' @param \dots not used
#' @param drop logical(1) set TRUE to drop array character 
setMethod("[", c("H5S_dataset", "character", "character", "ANY"), function(x, i, j, ..., drop=FALSE) {

  uu = x@presel
  dims = x@shapes$dims

  ind1lims = as.numeric(strsplit(i, ":")[[1]])
  if (ind1lims[1] < 0) 
    stop("negative starting index not allowed in i")
  if (ind1lims[2] > dims[1]) 
    stop("i exceeds boundary for first index")

  ind2lims = as.numeric(strsplit(j, ":")[[1]])
  if (ind2lims[1] < 0) 
    stop("negative starting index not allowed in j")
  if (ind2lims[2] > dims[2]) 
    stop("j exceeds boundary for second index")

  delta1 <- 1
  if ( length(ind1lims) >= 3) {
    delta1 <- ind1lims[3]
  }
  nrow <- ceiling((ind1lims[2] - ind1lims[1])/delta1)  

  delta2 <- 1
  if ( length(ind2lims) >= 3) {
    delta2 <- ind2lims[3]
  }
  ncol <- ceiling((ind2lims[2] - ind2lims[1])/delta2)  

  # flip negative ranges because HDF5 doesn't allow them
  if ( delta1 < 0 )  {
    # TODO: leave delta1 < 0 for flag
    i <- paste0(ind1lims[2], ":", ind1lims[1], ":", -delta1)
  }
  if ( delta2 < 0 )  {
    # TODO: leave delta2 < 0 for flag
    j <- paste0(ind2lims[2], ":", ind2lims[1], ":", -delta2)
  }
  nele <- nrow*ncol    
  uu = sub("%%SEL1%%", i, uu)
  uu = sub("%%SEL2%%", j, uu)
  h5typ <- x@allatts$type$base
  if ( ( h5typ == "H5T_IEEE_I32LE" | h5typ == "H5T_IEEE_F64LE" ) & 
         x@transfermode == "binary" )  {
    # message(paste("binary transfer", sep=""))
    val <- bintransl(uu, nele, h5typ)
  }
  else  {
    # message(paste("JSON transfer", sep=""))
    val <- jsontransl(uu, nele)
  }
  mat <- matrix(val, nrow=nrow, ncol=ncol, byrow = TRUE, dimnames = NULL)
  # flip back negative ranges
  if (delta1 < 0)  {
    mat <- mat[c(nrow:1),,drop=FALSE]
  }
  if (delta2 < 0)  {
    mat <- mat[,c(ncol:1),drop=FALSE]
  }
  mat
})

#' Find a dataset on source from its name
#'
#' This function is deprecated and will be defunct in the next release.
#'
#' @name dataset
#' @param h5s instance of H5S_source
#' @param tag character string identifying a dataset
#' @return object of type H5S_dataset
#' @export
dataset = function(h5s, tag) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  dsns = dsmeta(h5s)[["dsnames"]] # mcols problem with [,"dsnames"]
  # find row of dsmeta DataFrame where tag is a substring of a dataset name
  # to allow substrings to be used for querying (e.g., "100k" for "neurons100k")
  hits = vapply(dsns, function(x) length(grep(tag, x))>0, logical(1))
  if (!any(hits)) 
    stop("tag not found in dsmeta(h5s)")
  if (sum(hits)>1) 
    warning("tag occurs in several groups, using first")
  fulldsn = dsns[[which(hits)]] #[[1]] # unlist; which(hits) is relevant group
  fulldsn = fulldsn[ grep(tag, fulldsn) ] # find the actual simple name matching substring in [[]]
  lin = links(h5s, which(hits))
  targs = targets(lin)
  targs = targs[grep(tag,targs)]

  if (length(targs)>1)  {
    cat("dataset tag does not identify a unique target. Found:", "\n")
    cat(targs, sep="\n")
    stop("please supply tag to identify single target.")
  }

  targ = sub(".host", "datasets?host", targs)
  uuid = transl(targ)$datasets 
  attrs = transl( sub("datasets", paste0("datasets/", uuid), targ ) )
  hrnm = vapply(attrs$hrefs, "[[", character(1), 2)
  hrval = vapply(attrs$hrefs, "[[", character(1), 1)
  ans = DataFrame(hrefName=hrnm, hrefValue=hrval)
  rownames(ans) = hrnm
  self = ans["self", "hrefValue"]
  prep = sub("\\?host=", "/value?host=", self)
  prep = paste0(prep, "&select=[%%SEL1%%,%%SEL2%%]")
 
  xjson = "JSON"
  new("H5S_dataset", source=h5s, simpleName=fulldsn,
    shapes=attrs$shape, hrefs=ans, allatts=attrs, presel=prep, transfermode=xjson)
}

#' acquire internal HDF5 dimension information for matrix
#'
#' This function is deprecated and will be defunct in the next release.
#'
#' @param h5d instance of H5S_dataset
#' @return vector with dimensions of dataset
#' @examples
#' \dontrun{
#' bigec2 = H5S_source(URL_h5serv())
#' tex <- bigec2[["tenx_100k_sorted"]]
#' internalDim(tex)
#' }
#' @export
internalDim = function(h5d) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  d = slot(h5d, "shapes")$dims
  c(intl.dim1=d[1], intl.dim2=d[2])
}


#' HSDS server get request accessor
#'
#' This function is deprecated and will be defunct in the next release.
#'
#' @param object H5S_source instance
#' @return a data frame with response
#' @examples 
#' \dontrun{
#' hsdsCon = H5S_source(URL_hsds()) # hsds server connection
#' hsdsInfo(hsdsCon)
#' }
#' @rdname hsdsInfo
#' @aliases hsdsInfo,H5S_source-method
#' @docType methods
#' @export hsdsInfo
#' @exportMethod hsdsInfo
setGeneric("hsdsInfo", function(object) standardGeneric("hsdsInfo"))
setMethod("hsdsInfo", c("H5S_source"), function(object) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  #target = paste0(.serverURL(object))
  target = paste0(.serverURL(object),"/domains")
  ans = transl(target) 
  
  if (length(ans$domains)<1)  { 
    stop("no domains at server URL")
  }
  hh = t(sapply(ans$domains, force)) 
  DataFrame(hh)
  #domains in the home folder
  #target = paste0(.serverURL(object), "/domains?domain=/home/")
  #ans = transl(target)
  #hd = t(sapply(ans$domains, force))
  #DataFrame(hd)
  
})

#' HSDS server domains accessor
#'
#' This function is deprecated and will be defunct in the next release.
#'
#' @param object H5S_source instance
#' @param \dots not used
#' @return a data frame with domains name
#' @examples 
#' \dontrun{
#' hsdsCon = H5S_source(URL_hsds()) # hsds server connection
#' setPath(hsdsCon, "/home/stvjc/")-> hsds
#' domains(hsds)
#' }
#' @rdname domains
#' @aliases domains,H5S_source-method
#' @docType methods
#' @export domains
#' @exportMethod domains
setGeneric("domains", function(object, ...) standardGeneric("domains"))
setMethod("domains", c("H5S_source"), function(object, ...) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  target = paste0(.serverURL(object, ...),"/domains?domain=", object@folderPath)
  #target = paste0(.serverURL(object))
  ans = transl(target) 
  
  if (length(ans$domains)<1)  { 
    stop("no domains at server URL")
  }
  hh = t(sapply(ans$domains, force)) 
  DataFrame(hh)
  #domains in the home folder
  #target = paste0(.serverURL(object), "/domains?domain=/home/")
  #ans = transl(target)
  #hd = t(sapply(ans$domains, force))
  #DataFrame(hd)
  
})

#'  getDatasetUUIDs from hsds server
#' 
#'  This function is deprecated and will be defunct in the next release.
#' 
#' @param object instance of H5S_source(updated object with path to file set)
#' @return character of dataset uuid obtained 
#' @examples
#' \dontrun{
#' hsdsCon = H5S_source(URL_hsds()) # hsds server
#' setPath(hsdsCon, "/home/stvjc/hdf5_mat.h5")-> hsds
#' getDatasetUUIDs(hsds)
#' }
#' @rdname getDatasetUUIDs
#' @export
getDatasetUUIDs <- function(object) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  query = sprintf("%s/datasets?host=%s", object@serverURL, object@folderPath)
  if (.Platform$OS.type == "windows") query = winpref(query)
  ans = try(GET(query))
  if (inherits(ans, "try-error")) stop("could not resolve datasets query")
  cont = fromJSON(readBin( ans$content, what="character"))
  cont$datasets
}

#' getDatasetAttrs from hsds server
#' 
#'  This function is deprecated and will be defunct in the next release.
#' 
#' @param object instance of H5S_source(updated object with path to file set)
#' @param duid character string with dataset uuid
#' @return list of data obtained
#' @examples
#' \dontrun{ 
#' hsdsCon = H5S_source(URL_hsds()) # hsds server
#' hsdsCon@folderPath="/home/stvjc/hdf5_mat.h5"
#' ds = fetchDatasets(hsdsCon)# Pick the ID of the dataset you are interested in
#' getDatasetAttrs(hsdsCon, "d-a9e4b71c-8ea2-11e8-9306-0242ac120022")
#' }
#' @rdname getDatasetAttrs
#' @export
getDatasetAttrs <- function(object, duid) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  #uu = getDatasetUUIDs(object)
  uu = duid
  query = sprintf("%s/datasets/%s?host=%s", object@serverURL, uu, object@folderPath)
  if (.Platform$OS.type == "windows") query = winpref(query)
  ans = try(GET(query))   ## is this going to do GET request only for one url? what if the there are multiple datasets in the file?
  if (inherits(ans, "try-error")) stop("could not resolve datasets query")
  cont = fromJSON(readBin( ans$content, what="character"))
  cont
}

#' getDims from hsds server
#' 
#'  This function is deprecated and will be defunct in the next release.
#' 
#' @param object instance of H5S_source(updated object with path to file set)
#' @param duid character string with dataset uuid
#' @return numeric content of dimensions
#' @examples
#' \dontrun{
#' hsdsCon = H5S_source(URL_hsds()) # hsds server
#' setPath(hsdsCon, "/home/stvjc/hdf5_mat.h5")-> hsds
#' duid <- 'd-a9e4b71c-8ea2-11e8-9306-0242ac120022'
#' getDims(hsds, duid)
#' }
#' @rdname getDims
#' @export
getDims <- function(object, duid) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  stopifnot(is(object, "H5S_source"))
  getDatasetAttrs(object, duid)$shape$dims
}

#' getHRDF from hsds server
#' 
#'  This function is deprecated and will be defunct in the next release.
#' 
#' @param object instance of H5S_source(updated object with path to file set)
#' @param duid character string with dataset uuid
#' @return DataFrame of data obtained
#' @examples
#' \dontrun{
#' hsdsCon = H5S_source(URL_hsds()) # hsds server
#' hsdsCon@folderPath="/home/stvjc/hdf5_mat.h5"
#' ds = fetchDatasets(hsdsCon) #Pick the ID of the dataset you are interested in
#' getHRDF(hsdsCon, "d-a9e4b71c-8ea2-11e8-9306-0242ac120022")
#' }
#' @rdname getHRDF
#' @export
getHRDF <- function(object, duid) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  stopifnot(is(object, "H5S_source"))
  atts = getDatasetAttrs(object, duid)
  nms = sapply(atts$hrefs, "[[", "rel")
  vals = sapply(atts$hrefs, "[[", "href")
  DataFrame(hrefName=nms, hrefValue=vals)
}

#' H5S_dataset2 for datasets in hsds server
#' 
#'  This function is deprecated and will be defunct in the next release.
#' 
#' @param object instance of H5S_source(updated object with path to file set)
#' @param duid character vector with dataset uuid of interest
#' @return H5S_dataset object
#' @examples
#' \dontrun{
#' hsdsCon = H5S_source(URL_hsds()) # hsds server
#' hsdsCon@folderPath="/home/stvjc/hdf5_mat.h5"
#' ds = fetchDatasets(hsdsCon) #Pick the dataset id of interest
#' H5S_dataset2(hsdsCon, "d-a9e4b71c-8ea2-11e8-9306-0242ac120022")
#' }
#' @rdname H5S_dataset2
#' @export
H5S_dataset2 = function(object, duid) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  src = new("H5S_source", serverURL=object@serverURL, dsmeta=DataFrame())
  atts = getDatasetAttrs(object, duid)
  ans = getHRDF(object, duid)
  rownames(ans) = ans[,1]
  self = ans["self", "hrefValue"]
  prep = sub("\\?host=", "/value?host=", self)
  prep = paste0(prep, "&select=[%%SEL1%%,%%SEL2%%]")
  url = paste0(object@serverURL,"/datasets/",duid,"?host=",object@folderPath)
  if (.Platform$OS.type == "windows") url = winpref(url)
  res = fromJSON(content(GET(url),"text"))
  new("H5S_dataset", source=src, simpleName=object@folderPath, shapes=res$shape,
      hrefs = ans, allatts=atts, presel=prep, transfermode="JSON")
}

#' getDatasetSlice from hsds server
#' 
#'  This function is deprecated and will be defunct in the next release.
#' 
#' @param object instance of H5S_source(updated object with path to file set)
#' @param dsindex dataset index
#' @param selectionString character with selectionString
#' @param \dots unused
#' @return list of data obtained
#' @examples
#' \dontrun{
#' hsdsCon = H5S_source(URL_hsds()) # hsds server
#' setPath(hsdsCon, "/home/stvjc/hdf5_mat.h5")-> hsds
#' getDatasetSlice(hsds,dsindex=1,selectionString="[1:2,1:5]")
#' }
#'@export
getDatasetSlice <- function(object, dsindex=1, selectionString, ...) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  requireNamespace("httr")
  requireNamespace("rjson")
  uuid = getDatasetUUIDs(object)[dsindex]
  query = sprintf("%s/datasets/%s/value?host=%s&select=%s", object@serverURL, uuid, object@folderPath, selectionString)
  if (.Platform$OS.type == "windows") query = winpref(query)
  ans = try(GET(query, add_headers(Accept="application/json")))
  if (inherits(ans, "try-error")) stop("could not resolve select query")
  fromJSON(readBin( ans$content, what="character"))
}

#' fetch datasets of a hdf5 file from the hsds server
#' 
#'  This function is deprecated and will be defunct in the next release.
#' 
#' @import R6
#' @import httr
#' @param object instance of H5S_source
#' @return data.frame with information about the datasets in the file 
#' @examples
#' \dontrun{  
#' hsdsCon = H5S_source(URL_hsds()) # hsds server
#' hsdsCon@folderPath="/home/stvjc/hdf5_mat.h5"
#' ds = fetchDatasets(hsdsCon)
#' ds
#' }
#' @export
fetchDatasets <- function(object){
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  hdf5DataStore <- R6Class("hdf5DataStore", list(
    serverURL="",
    domain="",
    rootgroup="",
    initialize=function(serverURL=NA, domain=NA){
      self$serverURL=serverURL;
      self$domain=domain;
      if (.Platform$OS.type == "windows") serverURL = winpref(serverURL)
      temp=fromJSON(content(GET(paste0(serverURL,"/domains?domain=",domain)),"text"))
      self$rootgroup=temp$domains[[1]]$root
      self$updateDatasets(temp$domains[[1]]$root, domain)
    },
    datasets=data.frame(),
    updateDatasets=function(group, domain){
      url=paste0(self$serverURL,"/groups/",group,"/links?domain=",domain)
      if (.Platform$OS.type == "windows") url = winpref(url)
      a=fromJSON(content(GET(url),"text"))$links
      for(i in 1:length(a)){
        a[[i]]$domain=domain
        if("collection" %in% names(a[[i]])){
          if(a[[i]]$collection=="datasets"){
            if(nrow(self$datasets)<1){
              self$datasets=as.data.frame(a[[i]],stringsAsFactors=FALSE)}else{
                self$datasets=rbind(self$datasets,
                                    as.data.frame(a[[i]],stringsAsFactors=FALSE))}
          }else{
            self$updateDatasets(a[[i]]$id,domain)
          }
        }
      }
    }
  ))
  myDS=hdf5DataStore$new(object@serverURL, object@folderPath)
  myDS$datasets
}

# 30 August 2018 -- purpose of this code is
# to verify that key subsetting code related to
# H5S_dataset2 does not regress
#
getDSrefs = function(url = rhdf5client::URL_hsds(), 
   dompath = "/shared/bioconductor/htxcomp_genes.h5") {
 ini = httr::GET(paste0(url, "/datasets?domain=", dompath))
 val = try(rjson::fromJSON(readBin(ini$content, "text")))
 if (inherits(val, "try-error")) {
   print(val)
   stop("GET /datasets failed to produce useful content.")
   }
 groups = httr::GET(paste0(url, "/groups?domain=", dompath))
 groups.val = try(rjson::fromJSON(readBin(groups$content, "text")))
 if (inherits(groups.val, "try-error")) {
   print(val)
   stop("GET /groups failed to produce useful content.")
   }
 gtab = do.call(rbind.data.frame, c(groups.val$hrefs, stringsAsFactors=FALSE))
 rownames(gtab) = gtab$rel
 prelink = gtab["root","href"]
 lurl = sub("\\?", "/links?", prelink)
 links = httr::GET(lurl)
 links.val = try(rjson::fromJSON(readBin(links$content, "text")))
 if (inherits(links.val, "try-error")) {
   print(val)
   stop("GET /groups/.../links failed to produce useful content.")
   }
 ltab = do.call(rbind.data.frame, c(links.val$links, stringsAsFactors=FALSE))
 ltab
}

retrieveDataset = function(url=rhdf5client::URL_hsds(), dompath =
  "/shared/bioconductor/darmgcls.h5", title="assay001") {
#
# function will attempt to return a DelayedMatrix with
# contents given by the [title] element of the [dompath]
# resource, unless the [title] element is one-dimensional,
# in which case a character vector is returned
#
 s1 = getDSrefs(url, dompath=dompath)
 uu = s1[which(s1$title == title), "id"]
 so = rhdf5client::H5S_source(url, dompath)
 suppressMessages({  # encoding message
 ds = rhdf5client::H5S_dataset2(so, uu)
 })
 if (length(ds@shapes$dims) == 1) {
        targ = gsub("&.*", "", ds@presel)
        gg = rjson::fromJSON(readBin(httr::GET(targ)$content, w = "character"))$value
        message("one dimensional response, returning text")
        return(gg)
    }
 DelayedArray::DelayedArray(new("H5S_ArraySeed", filepath = "", domain = "", 
         host = "", H5S_dataset = ds))
}
 
context("H5S_dataset2 exercises")

test_that("darmgcls.h5 can be retrieved from kitalab", {
  if (check_hsds()) { 
    url = URL_hsds()
    dompath = "/shared/bioconductor/darmgcls.h5"
    title = "assay001"
    s1 = try(getDSrefs(url = url, dompath = dompath))
    if (inherits(s1, "try-error")) {
      print(s1)
      fail(paste("getDSrefs fails on", dompath))
    }
    uu = s1[which(s1$title == title), "id"]
    expect_true(length(uu)==1)
    so = rhdf5client::H5S_source(url, dompath)
    suppressMessages({  # encoding message
    ds = rhdf5client::H5S_dataset2(so, uu)
    })
    expect_true(is(ds, "H5S_dataset"))
    ans = DelayedArray::DelayedArray(new("H5S_ArraySeed", 
            filepath = "", domain = "", 
            host = "", H5S_dataset = ds))
    expect_true(all(dim(ans) == c(65218, 3584)))
   
   # compute indices of corners of matrix
   
   corn = function(m, siz=3) {
    r = nrow(m)
    c = ncol(m)
    rr = unlist(lapply(list(head,tail), function(f) f(1:r, siz)))
    cc = unlist(lapply(list(head,tail), function(f) f(1:c, siz)))
    list(rr, cc)
   }
   
   mcorn = function(m, siz=4) { 
     cc = corn(m, siz=siz)
     m[cc[[1]], cc[[2]]]
   }
   
   expect_equal(sum(mcorn(ans)), 55928.6253033)
   }
  else TRUE
})

context("use retrieveDataset to get rownames")

test_that("retrieveDataset gets rownames", {
  if (check_hsds()) {
   rn = retrieveDataset(url=rhdf5client::URL_hsds(), dompath =
       "/shared/bioconductor/htxcomp_genes.h5", title="rownames") 
   expect_equal(length(rn), 58288)
   expect_true(all.equal(head(rn), c("ENSG00000000003.14", "ENSG00000000005.5", 	"ENSG00000000419.12", "ENSG00000000457.13", 
        "ENSG00000000460.16", "ENSG00000000938.12")))
  } else TRUE
})

library(rhdf5client)
system("rm -f profile.out")
fname <- "profile.out"

bigec2 <- H5S_source("http://54.174.163.77:5000")
txdat <- bigec2[["tenx_100k_sorted"]]

cat("-------------------------------------------------------------------------",
  file=fname, sep="\n", append=TRUE)
cat("2000 x 2000 JSON transfer", file=fname, sep="\n", append=TRUE)
Rprof()
for (i in 1:5)  {
  M <- submatrix(txdat, 1:2000, 1:2000)
}
Rprof(NULL)
out <- capture.output(summaryRprof())
cat(out, file=fname, sep="\n", append=TRUE)

cat("-------------------------------------------------------------------------",
  file=fname, sep="\n", append=TRUE)
transfermode(txdat) <- "binary"
cat("2000 x 2000 binary transfer", file=fname, sep="\n", append=TRUE)
Rprof()
for (i in 1:5)  {
  M <- submatrix(txdat, 1:2000, 1:2000)
}
Rprof(NULL)
out <- capture.output(summaryRprof())
cat(out, file=fname, sep="\n", append=TRUE)

cat("-------------------------------------------------------------------------",
  file=fname, sep="\n", append=TRUE)
transfermode(txdat) <- "JSON"

cat("20000 x 200 JSON transfer", file=fname, sep="\n", append=TRUE)
Rprof()
for (i in 1:5)  {
  M <- submatrix(txdat, 1:20000, 1:200)
}
Rprof(NULL)
out <- capture.output(summaryRprof())
cat(out, file=fname, sep="\n", append=TRUE)

cat("-------------------------------------------------------------------------",
  file=fname, sep="\n", append=TRUE)
transfermode(txdat) <- "binary"
cat("20000 x 200 binary transfer", file=fname, sep="\n", append=TRUE)
Rprof()
for (i in 1:5)  {
  M <- submatrix(txdat, 1:20000, 1:200)
}
Rprof(NULL)
out <- capture.output(summaryRprof())
cat(out, file=fname, sep="\n", append=TRUE)

cat("-------------------------------------------------------------------------",
  file=fname, sep="\n", append=TRUE)
transfermode(txdat) <- "JSON"

cat("200 x 20000 JSON transfer", file=fname, sep="\n", append=TRUE)
Rprof()
for (i in 1:5)  {
  M <- submatrix(txdat, 1:200, 1:20000)
}
Rprof(NULL)
out <- capture.output(summaryRprof())
cat(out, file=fname, sep="\n", append=TRUE)

cat("-------------------------------------------------------------------------",
  file=fname, sep="\n", append=TRUE)
transfermode(txdat) <- "binary"
cat("200 x 20000 binary transfer", file=fname, sep="\n", append=TRUE)
Rprof()
for (i in 1:5)  {
  M <- submatrix(txdat, 1:200, 1:20000)
}
Rprof(NULL)
out <- capture.output(summaryRprof())
cat(out, file=fname, sep="\n", append=TRUE)

# 
# implementation of 
# https://github.com/Bioconductor/DelayedArray/blob/master/vignettes/02-Implementing_a_backend.Rmd
# for HDF server back end, based on rhdf5client
#
# key contract: if is (X, "H5S_source") and host is the
# name of an HDF5 file on the server, in /data, then Z = X[[ host ]]
# returns an H5S_dataset instance; this in turn answers to Z[i,j]
# with a numeric matrix
#
# longrunning jan 2018

deprecate_msg = paste0("This function is deprecated. The new interface to rhdf5client",
 " is exclusively through its DelayedArray backend HSDSArray")

# (legacy) with h5serv omit filepath and look in dsmeta for dataset
# endpoint URL_h5serv()
# filepath ''
# host tenx_full
# With hsds pass endpoint, filepath and dataset name
# endpoint URL_hsls()
# filepath '/home/spollack/testzero.h5'
# host 'dsetAA1'

#' H5S_Array for HDF Server content
#' @import DelayedArray
setClass("H5S_ArraySeed",
   contains="Array",
   slots = c(
     endpoint="character",
     filepath="character",
     domain="character",
     host="character",
     H5S_dataset="H5S_dataset"
     ))

H5S_ArraySeed = function(endpoint, filepath, host) {
  requireNamespace("rhdf5client")
  con <- try(rhdf5client::H5S_source(endpoint))

  if (!is(con, "H5S_source")) stop("could not resolve H5S_source request on filepath")
#
# for HDF Server back end, the following invocation of [[
# establishes the link to numerical data source
#

# detect h5serv/hsds by checking for dsmeta as in H5S_source constructor 
# TODO: revisit this decision

  if (nrow(con@dsmeta)>1)  {  # h5serv
    ds <- try(con[[host]])
  } else  {   # hsds
    con <- setPath(con, filepath)
    df <- fetchDatasets(con)
    ids <- df$id[which(df$title==host)]
    if (length(ids) == 1)  {   
      uuid <- ids[1]
      ds <- H5S_dataset2(con, uuid)
    } else if (length(ids) == 0)  {
      stop(paste0("no such dataset ",host," in file ",filepath))
    } else  {
      stop(paste0("multiple datasets ",host," in file ",filepath))
    }
  }

  if (!is(ds, "H5S_dataset")) stop("could not resolve H5S_dataset request on filepath[[host]]")
  new("H5S_ArraySeed", endpoint=endpoint, 
       domain=filepath, host=host, H5S_dataset=ds)
  }

#' dimnames not stored with H5S_source as of Jan 2018
#'
#' This function is deprecated and will be defunct in the next release.
#'
#' @param x instance of H5S_ArraySeed
#' @return currently returns list(NULL, NULL) as we do not store dimnames in HDF5
#' @export
setMethod("dimnames", "H5S_ArraySeed", function(x) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  list(NULL, NULL)
})

#' HDF Server content is assumed transposed relative to R matrix layout
#'
#' This function is deprecated and will be defunct in the next release.
#'
#' @param x instance of H5S_ArraySeed
#' @return integer(2) vector of dimensions corresponding to R's layout, assuming 2-d data
#' @export
setMethod("dim", "H5S_ArraySeed", function(x) {
  # note that for HDF Server the internal dims are
  # transposed relative to R expectations
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  rev(as.integer(x@H5S_dataset@shapes$dims))
})

#
# this seems incredibly convoluted 
#
setMethod("extract_array", "H5S_ArraySeed", function(x, index) {
  stopifnot(length(index)==2)
  dims_at_source = rev(dim(x))
  index = rev(index)
  full_inds_for_H5S = list(as.integer(seq_len(dims_at_source[[1]])),
        as.integer(seq_len(dims_at_source[[2]])))
  request = vector("list", 2)
# the request will operate in transpose space of R index
  if (is.null(index[[1]])) request[[2]] = full_inds_for_H5S[[1]]
    else if (length(index[[1]]) == 0) request[2] = list(NULL)
    else request[[2]] = index[[1]]
  if (is.null(index[[2]])) request[[1]] = full_inds_for_H5S[[2]]
    else if (length(index[[2]]) == 0) request[1] = list(NULL)
    else request[[1]] = index[[2]]
  reqcl = unlist(lapply(request, is.null))
  if (any(reqcl)) { # will have a zero dimension, can return without query
    if (all(reqcl)) return(matrix(numeric(0), 0, 0))
    else if (reqcl[1]) return(t(matrix(numeric(0), length(request[[2]]), 0)))
    else return(t(matrix(numeric(0), 0, length(request[[1]]))))
    }
  if (max(request[[1]]) > dims_at_source[2]) stop("dim1 request out of bounds")
  if (max(request[[2]]) > dims_at_source[1]) stop("dim2 request out of bounds")
  t(x@H5S_dataset[ request[[2]], request[[1]] ]) # native method has drop=FALSE
})
  
#' extension of DelayedArray for HDF Server content
#' @exportClass H5S_Array
setClass("H5S_Array", contains="DelayedArray")

#' extension of DelayedMatrix for HDF Server content
#' @exportClass H5S_Matrix
setClass("H5S_Matrix", contains=c("DelayedMatrix", 
     "H5S_Array"))

setMethod("matrixClass", "H5S_Array", function(x) "H5S_Matrix")

#' coercion for remote array to remote matrix
#' @name as
#' @family H5S_Array

setAs("H5S_Array", "H5S_Matrix", function(from)
   new("H5S_Matrix", from))

setMethod("DelayedArray", "H5S_ArraySeed",
   function(seed) new_DelayedArray(seed, Class="H5S_Array"))

#' create H5S_Array instance given url (filepath) and entity (host) name
#'
#' This function is deprecated and will be defunct in the next release.
#'
#' @param endpoint a character(1) URL to port for HDF Server
#' @param filepath path and name of the H5 file
#' @param host a character(1) name of 'host' in server
#' @return an instance of \code{\link[DelayedArray]{DelayedArray-class}}
#' @examples
#' # The true values from yriMulti data element 'banovichSE':
#' # > assay(banovichSE[c(1:5,329465:329469),c(1:3,63:64)])
#' #                    NA18498    NA18499    NA18501  |    NA18489     NA18909
#' # cg00000029      0.47339629  1.2943041 -0.8084735  |  0.6708168 -0.86093022
#' # cg00000165      1.23640861  0.2099817 -0.2683763  |  0.4446088  0.99868231
#' # cg00000236     -0.22258183  1.6236857 -0.8654838  |  0.1958195 -0.06090929
#' # cg00000289      0.65720581  0.5527470 -1.8458295  | -0.4618782  0.34934164
#' # cg00000363     -0.15063083  0.7498020  0.3254333  |  0.7342878  0.12940774
#' # #-------------------------------------------------------------------------
#' # ch.9.98936572R -0.07954958  0.2139431 -0.4719621  |  0.6835012  0.57758798
#' # ch.9.98937537R  0.04254705  1.0702770  1.7356387  | -0.1531732 -1.52889773
#' # ch.9.98959675F -1.59253143  0.2982456 -1.1954030  | -1.3703135  0.28974909
#' # ch.9.98989607R -1.80646652  0.4760022  1.4771808  |  0.9479602  0.49921375
#' # ch.9.991104F    0.08180195 -0.2434306  1.0281002  | -0.1653721  0.55612215
#' #
#' @export
H5S_Array = function(endpoint, filepath, host)   {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  DelayedArray(H5S_ArraySeed(endpoint, filepath, host))
}

deprecate_msg = paste0("This function is deprecated. The new interface to rhdf5client",
 " is exclusively through its DelayedArray backend HSDSArray")

#' simplify construction of DelayedMatrix from url and path in HSDS
#'
#' This class is deprecated and will be defunct in the next release.
#'
#' @param url character(1) URL for HSDS object store with port
#' @param path character(1) path from root defining HDF Cloud resource
#' @return instance of DelayedArray
#' @examples
#' \dontrun{
#' HSDS_Matrix
#' }
#' @export
HSDS_Matrix_OLD = function(url, path) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  so = H5S_source(url, path)
  dss = fetchDatasets(so) 
  uu = dss$id
  if (length(uu)>1) message("multiple UUIDs found, using first")
  uu = uu[1]
  ds = H5S_dataset2(so, uu)
  DelayedArray(new("H5S_ArraySeed", filepath="", domain="",
   host="", H5S_dataset=ds))
}

#' simplify construction of DelayedMatrix from url and path in HSDS
#'
#' This class is deprecated and will be defunct in the next release.
#'
#' @param url character(1) URL for HSDS object store with port
#' @param path character(1) path from root defining HDF Cloud resource
#' @param title character(1) name of dataset to use
#' @return instance of DelayedArray
#' @examples
#' \dontrun{
#' HSDS_Matrix(URL_hsds(), "/shared/bioconductor/darmgcls.h5")
#' }
#' @export
HSDS_Matrix = function(url, path, title) {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  so = H5S_source(url, path)
  dss = fetchDatasets(so)
  uu = dss$id
  if (length(uu)>1) {
    message("multiple UUIDs found")
    if (missing(title)) {
      message("no title provided, using first UUID")
      uu = uu[1]
      }
    else {
      ind = which(dss$title == title)
      if (length(ind)==1) uu = uu[ind]
        else stop("title does not pick out a single dataset to use")
      }
    }
  ds = H5S_dataset2(so, uu)
  if (length(ds@shapes$dims)==1) {
       targ = gsub("&.*", "", ds@presel)
       gg = fromJSON(readBin(GET(targ)$content,what="character"))$value
       message("one dimensional response, returning text")
       return(gg)
       }
  DelayedArray(new("H5S_ArraySeed", filepath="", domain="",
   host="", H5S_dataset=ds))
}

deprecate_msg = paste0("This function is deprecated. The new interface to rhdf5client",
 " is exclusively through its DelayedArray backend HSDSArray")

#' An S4 class to represent a dataset in a HDF5 file.
#' @import httr methods rjson
#' @slot file An object of type HSDSFile; the file in which the dataset is resident.
#' @slot path The dataset's path in the internal HDF5 hiearchy.
#' @slot uuid The unique unit ID by which the dataset is accessed in the server 
#' database system.
#' @slot shape The dimensions of the dataset
#' @slot type The dataset's HDF5 datatype
setClass("HSDSDataset", representation(file="HSDSFile", path="character", uuid="character",
  shape="numeric", type="list"))

#' Construct an object of type HSDSDataset 
#' A HSDSDataset is a representation of a dataset in a HDF5 file.
#' @name HSDSDataset
#' @param file An object of type HSDSFile which hosts the dataset 
#' @param path The complete intrafile path to the dataset
#' @return An initialized object of type HSDSDataset
#' @examples
#' if (check_hsds()) {
#'  src <- HSDSSource(URL_hsds())
#'  f <- HSDSFile(src, '/shared/bioconductor/patelGBMSC.h5')
#'  d <- HSDSDataset(f, '/assay001')
#' }
#' @export
HSDSDataset <- function(file, path)  {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  idx <- which(file@dsetdf[,1] == path)
  if (length(idx) == 0)  
    stop("no such dataset")
  uuid <- file@dsetdf[idx,2]
  request <- paste0(file@src@endpoint, '/datasets/', uuid, '?domain=', file@domain)
  response <- tryCatch(submitRequest(request),
    error=function(e) { NULL })
  if (is.null(response))  { # this should be almost impossible
    warning("bad http request", call. = FALSE)
    return(NULL)
  }
  
  if (response$shape$class == "H5S_SCALAR") {
    shape <- 1
  } else {
    shape <- response$shape$dims
  }
  
  type <- response$type

  new("HSDSDataset", file=file, path=path, uuid=uuid,
    shape=shape, type=type)
}

.HSDSDataset <- function(file, path)  { # after deprecation cycle this is all that will remain
  idx <- which(file@dsetdf[,1] == path)
  if (length(idx) == 0)  
    stop("no such dataset")
  uuid <- file@dsetdf[idx,2]
  request <- paste0(file@src@endpoint, '/datasets/', uuid, '?domain=', file@domain)
  response <- tryCatch(submitRequest(request),
    error=function(e) { NULL })
  if (is.null(response))  { # this should be almost impossible
    warning("bad http request", call. = FALSE)
    return(NULL)
  }
  shape <- response$shape$dims
  type <- response$type

  new("HSDSDataset", file=file, path=path, uuid=uuid,
    shape=shape, type=type)
}

#' extract elements of a one or two-dimensional HSDSDataset
#' 
#' Fetch data from a remote dataset
#'
#'
#' The servers require data to be fetched in slices, i.e., in sets of 
#' for which the indices of each dimension are of the form start:stop:step.
#' More complex sets of indices will be split into slices and fetched in
#' multiple requests. This is opaque to the user, but may enter into 
#' considerations of data access patterns, e.g., for performance-tuning.
#'
#' @param dataset An object of type HSDSDataset, the dataset to access.
#'
#' @param indices The indices of the data to fetch
#'
#' @param transfermode Either 'JSON' or 'binary' (default)
#'
#' @return an Array containing the data fetched from the server
#'
#' @docType methods
#' @rdname getData-methods
#'
#' @examples
#' if (check_hsds()) {
#'  s <- HSDSSource(URL_hsds())
#'  f <- HSDSFile(s, '/shared/bioconductor/patelGBMSC.h5')
#'  d <- HSDSDataset(f, '/assay001')
#'  x <- getData(d, c('1:4', '1:27998'), transfermode='JSON')
#'  xb <- getData(d, c('1:4', '1:27998'), transfermode='binary')
#'  # x <- getData(d, c(1:4, 1:27998), transfermode='JSON') # method missing?
#'  x 
#'  xb
#' }
#' @export 
setGeneric("getData", function(dataset, indices, transfermode) standardGeneric("getData"))
#
# need private version after deprecation cycle
#
setGeneric(".getData", function(dataset, indices, transfermode) standardGeneric(".getData"))

#' @rdname getData-methods
#' @aliases getData,HSDSDataset,character,character-method
setMethod("getData", c("HSDSDataset", "character", "character"),  
  function(dataset, indices, transfermode)  {
    #.Deprecated("HSDSArray", NULL, deprecate_msg)
    getDataVec(dataset, indices, transfermode)
  })

#private for after cycle
setMethod(".getData", c("HSDSDataset", "character", "character"),  
  function(dataset, indices, transfermode)  {
    getDataVec(dataset, indices, transfermode)
  })

#' @rdname getData-methods
#' @aliases getData,HSDSDataset,character,missing-method
setMethod("getData", c("HSDSDataset", "character", "missing"),  
  function(dataset, indices)  {
    #.Deprecated("HSDSArray", NULL, deprecate_msg)
    getDataVec(dataset, indices, 'binary')
  })
#private for after cycle
setMethod(".getData", c("HSDSDataset", "character", "missing"),  
  function(dataset, indices)  {
    getDataVec(dataset, indices, 'binary')
  })

#' @rdname getData-methods
#' @aliases getData,HSDSDataset,list,character-method
setMethod("getData", c("HSDSDataset", "list", "character"),  
function(dataset, indices, transfermode)  {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  getDataList(dataset, indices, transfermode)  
  })

#private for after cycle
setMethod(".getData", c("HSDSDataset", "list", "character"),  
function(dataset, indices, transfermode)  {
  getDataList(dataset, indices, transfermode)  
  })

#' @rdname getData-methods
#' @aliases getData,HSDSDataset,list,,missing-method
setMethod("getData", c("HSDSDataset", "list", "missing"),  
function(dataset, indices)  {
  #.Deprecated("HSDSArray", NULL, deprecate_msg)
  getDataList(dataset, indices, 'binary')  
  })


# private - perform a single fetch; indices is a vector of
# type character with one slice per dimension.
#' @useDynLib rhdf5client, .registration = TRUE
getDataVec <- function(dataset, indices, transfermode = 'JSON', returnRaw=FALSE)  {

    indices <- checkSlices(dataset@shape, indices)
    if (length(indices) == 0)
      stop("bad slices")
    if (!(transfermode %in% c('JSON', 'binary')))  {
      warning('unrecognized transfermode, using JSON')
      transfermode <- 'JSON'
    }

    sdims <- vapply(indices, slicelen, numeric(1))

    # rdims is dimensions of response$value
    singlefetch <- FALSE
    if (all(sdims == 1))  {
      # this is a single-value fetch
      rdims <- c(1)
      singlefetch <- TRUE
    } else  {
      # h5pyd drops single-width dimensions at the end. 
      # dimensions of the result
      rdims <- sdims[which(sdims != 1)]
      while(sdims[length(sdims)] == 1)
        sdims <- sdims[1:(length(sdims)-1)]
    }

    indices <- vapply(indices, function(s) { s }, character(1))
    sel <- paste0('[', paste(indices, collapse=','), ']')
    endpoint <- dataset@file@src@endpoint
    domain <- dataset@file@domain

    request <- paste0(endpoint, '/datasets/', dataset@uuid, 
      '/value?domain=', domain, '&select=', sel)
    response <- submitRequest(request, transfermode=transfermode)
    
    if (returnRaw) {
      return(response$value)
    }

#Browse[2]> response$headers$`content-type`
#[1] "application/octet-stream"

    if (singlefetch)  {  
      if (identical(response$headers$`content-type`, "application/octet-stream"))
        return(extractBinary(dataset@type, prod(rdims), response))
      
      if (dataset@type$class == "H5T_STRING") {
        # returning value as is
        return(response$value)
      } else if (dataset@type$class == "H5T_COMPOUND") {
        return(extractCompoundJSON(dataset@type, response$value))
      }
      
      return(as.numeric(response$value))  # this seems to assume JSON noted 2023.01.10
    }

    if (length(rdims) == 1 && length(sdims) == 1)  {    # 1D array quick bypass
      if (identical(response$headers$`content-type`, "application/octet-stream"))
        return(extractBinary(dataset@type, prod(rdims), response))
      
      if (dataset@type$class == "H5T_STRING") {
        # returning value as is
        return(response$value)
      } else if (dataset@type$class == "H5T_COMPOUND") {
        return(extractCompoundJSON(dataset@type, response$value))
      }
      
      return(as.numeric(response$value))  # loses array character, is that OK? 2023.01.10
    }

    if (length(rdims) == 2 && length(sdims) == 2)  {    # 2D array quick bypass
      if (transfermode == 'JSON')  {
        result <- response$value 
        A <- matrix(nrow = rdims[1], ncol = rdims[2])
        for (i in 1:rdims[1])  {
          A[i,] <- as.numeric(result[[i]])
        }
        return(A)
      } else if (transfermode == 'binary')  {
        result <- extractBinary(dataset@type, prod(rdims), response)
        A <- matrix(data=result, nrow = rdims[2], ncol = rdims[1])
        return(t(A))
      }
    }

    nn <- prod(rdims)
    A <- array(rep(0, nn))
    if (transfermode == 'JSON')  {
      # unpack response into an R array 
      result <- response$value

      # multi-dimensional data is returned as a list.
      if (is.list(result))  {
        A[1:nn] <- .Call("extractJSON", as.integer(length(sdims)), as.integer(sdims), result)
      } else  {  # one-dimensional data is returned as a vector 
        A[1:nn] <- as.numeric(result)
      }
    }  else if (transfermode == 'binary')  {
      result <- extractBinary(dataset@type, nn, response)
      A[1:nn] <- .Call("extractBin", as.integer(length(sdims)), as.integer(sdims), result)
    }

    # Question: should AA be forced to rdims or sdims?
    # For a slice of a multi-dimensional array, we want
    # to force to the subset. (rdims)

    # But if DelayedArray requires a 1 x N array, we
    # want the returned value to be 1 x N, not a 
    # vector of N. (sdims)

    # Arbitrary choice: force flat (rdims) and let 
    # the calling routine redimension.

    AA <- array(A, dim = rdims)
    AA
}

# private - split numeric vectors into slices and fetch
# data in one or more requests.
getDataList <- function(dataset, indices, transfermode = 'binary')  {
    if (length(dataset@shape) != length(indices))
      stop("wrong length of indexlist")

    slicelist <- lapply(indices, slicify)
    slclen <- lapply(slicelist, length)

    if (any(unlist(slclen) > 1))  {
      # assemble block arrays 
      AA <- multifetch(slicelist, dataset)
    } else  {
      # simple case: one block
      AA <- getDataVec(dataset, unlist(slicelist), transfermode)
    }
    AA

}

# private - in which we try to anticipate all the invalid things 
# users will try to enter for indices
checkSlices <- function(shape, slices)  {
  ok <- TRUE
  if (length(slices) != length(shape)) {
    message("wrong number of indices")
    ok <- FALSE
  }
  
  slicelist <- vector("list", length(slices))

  for (i in seq_along(slices))  {
    slice <- slices[i]
    start <- -1
    stop <- -1
    step <- -1

    st <- strsplit(slice, ':')[[1]]
    ss <- as.numeric(st)
    ss[which(is.na(ss))] <- -1

    if (slice == ':')  {
      start <- 1
      stop <- shape[i]
      step <- 1
    } else {
      if (length(ss) == 1)  {    # this is a slice like '5:' 
        start <- ss[1]
        stop <- shape[i]
        step <- 1
      } else if (length(ss) == 2)  {
        if (st[1] == '')  {      # ':5'
          start <- 1 
          stop <- ss[2]
          step <- 1
        } else  {
          start <- ss[1]
          stop <- ss[2]
          step <- 1
        }
      } else if (length(ss) == 3)  {
          start <- ss[1]
          stop <- ss[2]
          step <- ss[3]
      } else  {
        message(paste0("malformed slice ", i))
        ok <- FALSE
      }
    }

    slicevec <- c(start, stop, step)
    if (any(slicevec < 0))  {
      message(paste0("malformed slice ", i))
      ok <- FALSE
    }
    if (any(slicevec != trunc(slicevec)))  {
      message(paste0("malformed slice ", i))
      ok <- FALSE
    }

    slicelist[[i]] <- slicevec

    if (0 >= start || start > shape[i])  {
      message(paste0("slice start out of range in slice ", i))
      ok <- FALSE
    }
    if (0 >= stop || stop > shape[i])  {
      message(paste0("slice stop out of range in slice ", i))
      ok <- FALSE
    }
    if (stop < start)  {
      message(paste0("slice stop less than slice start in slice ", i))
      ok <- FALSE
    }
  }

  if (!ok)
    return(list())

  strslices <- lapply(slicelist, 
    function(slc) { sprintf('%d:%d:%d', slc[1], slc[2], slc[3]) })
  pyslices <- lapply(strslices, r2pyslice) 

}

# private - convert R-slice (string) to python-slice (string)
r2pyslice <- function(slcstr)  {
  slc <- as.numeric(strsplit(slcstr, ':')[[1]])
  # python indices from 0 instead of 1
  slc[1] <- slc[1]-1
  slc[2] <- slc[2]-1
  # if (stop-start) % step == 0, python slice excludes stop
  if ( (slc[2]-slc[1]) %% slc[3] == 0 )
    slc[2] <- slc[2] + 1
  slcstr <- sprintf('%d:%d:%d', slc[1], slc[2], slc[3])
}

# private - convert R-slice (vec) to python-slice (vec)
r2pyslcvec <- function(slc)  {
  # python indices from 0 instead of 1
  slc[1] <- slc[1]-1
  slc[2] <- slc[2]-1
  # if (stop-start) % step == 0, python slice excludes stop
  if ( (slc[2]-slc[1]) %% slc[3] == 0 )
    slc[2] <- slc[2] + 1
  slc
}

# Note: for more than two dimensions, "column-major" means
# "first-fastest" and "row-major" means "last-fastest"

# private - extract column-major subscripts for linear index
csub4idx <- function(D, ind)  {
  m <- length(D)
  X <- rep(-1, m)
  off <- ind-1
  
  X <- vapply(0:(m-1), function(j)  {
    if (j == 0)  {
      s <- off %% D[1]      
    } else if (j == m-1)  {
      s <- off %/% prod(D[1:m-1])
    } else  {
      s <- (off %/% prod(D[1:j])) %% D[j+1]
    } 
    s
  }, numeric(1))
  X <- X+1
  X
}

# private - extract column-major subscripts for linear index
rsub4idx <- function(D, ind)  {
  m <- length(D)
  X <- rep(-1, m)
  off <- ind-1
  
  X <- vapply(0:(m-1), function(j)  {
    if (j == 0)  {
      s <- off %/% prod(D[2:m])
    } else if (j == (m-1))  {
      s <- off %% D[m]
    } else  {
      s <- (off %/% prod(D[(j+2):m])) %% D[j+1]
    }
    s
  }, numeric(1))
  X <- X+1
  X
}

# private - extract column-major linear index for subscripts
cidx4sub <- function(D, S)  {
  m <- length(D)
  S <- S-1
  off <- 0
  for (j in 0:(m-1))  {
    p <- ifelse(j == 0, 1, prod(D[1:j]))
    off <- off + S[j+1]*p
  }
  off+1
}

# private - extract row-major linear index for subscripts
ridx4sub <- function(D, S)  {
  m <- length(D)
  S <- S-1
  off <- 0
  for (j in 0:(m-1))  {
    p <- ifelse(j == m-1, 1, prod(D[(j+2):m]))
    off <- off + S[j+1]*p
  }
  off+1
}



# private - length of valid (Python) slice
slicelen <- function(slc)  {
  ss <- as.numeric(strsplit(slc, ':')[[1]])
  sdim <- (ss[2]-ss[1]) %/% ss[3]
  if ((ss[2]-ss[1]) %% ss[3] != 0)
    sdim <- sdim + 1
  sdim
}


# private - extract binary data from response
# reference: https://support.hdfgroup.org/HDF5/doc1.8/RM/PredefDTypes.html
#' @importFrom utils strcapture
extractBinary <- function(typ, nele, rsp)  {

  # standard defaults
  what <- 'integer'
  size <- NA_integer_
  signed <- TRUE
  endian <- .Platform$endian     # just a guess - the server determines the endianness

  df <- strcapture('H5T_([[:alnum:]]*)_([IFUBD])([[:digit:]]+)([LB]E)', typ$base, 
                   data.frame(cl=character(), wh=character(), sz=integer(), 
                              en=character(), stringsAsFactors=FALSE))
  if (nrow(df) != 1) 
    stop(paste0("binary transfer for type ", typ$base, " not implemented yet"))
  if (!(df[1,1] %in% c('STD', 'IEEE')))
    stop(paste0("binary transfer for type ", typ$base, " not implemented yet"))

  if (df[1,2] == 'I' && df[1,3] == '32')  {
    what <- 'integer'
    size <- 4
  } else if (df[1,2] == 'I' && df[1,3] == '64')  {
    what <- 'integer'
    size <- 8
  } else if (df[1,2] == 'U' && df[1,3] == '32')  {
    what <- 'integer'
    size <- 4
    # not setting signed as it's ignored by readBin for this size
  } else if (df[1,2] == 'U' && df[1,3] == '64')  {
    what <- 'integer'
    size <- 8
    # not setting signed as it's ignored by readBin for this size
  } else if (df[1,2] == 'F' && df[1,3] == '64')  {
    what <- 'double'
    size <- 8
  } else  {
    stop(paste0("binary transfer for type ", typ$base, " not implemented yet"))
  }
  endian <- ifelse(df[1,4] == 'LE', 'little', 'big')
  result <- readBin(rsp$content, what=what, n=nele, size=size, signed=signed, endian=endian)
  result

}

# slicify - convert an arbitrary vector into slices
# This is an unsightly kludge, but it is designed to ensure
# that there is no more than one slice of width one.
# All other singletons should end up squashed into 
# pairs with each other. The idea is that the time 
# required to execute the loop is dwarfed by the time
# required to execute an extra remote fetch.
slicelst <- function(v)  {

  ll <- vector("list", length = length(v))   
  vec <- rep(0, length(v))
  
  il <- 1
  if (length(v) <= 2)  {
    ll <- list(v)
  } else {
  
    vec[1] <- v[1]
    vec[2] <- v[2]
    nv <- 2
    i <- 3
    while (i <= length(v))  {
      if (v[i]-v[i-1] == v[i-1]-v[i-2])  {
        nv <- nv + 1
        vec[nv] <- v[i]
        i = i + 1
      } else  {
        ll[[il]] <- vec[1:nv]
        il <- il + 1
        if (i < length(v))  {
          vec[1] <- v[i]
          vec[2] <- v[i+1]
          nv <- 2
          i = i + 2
        } else {
          vec[1] <- v[i]
          nv <- 1
          i = i + 1
        }
     }
    }
    if (nv > 0)  {
      ll[[il]] <- vec[1:nv]
    }
    ll <- ll[-which(sapply(ll, is.null))]
  }
  ll
}
  
# make list of vecs into string slices
slicify <- function(v)  {
  ll <- slicelst(v)
  slices <- vapply(ll, function(vec) { 
    start <- vec[1] 
    stop <- vec[length(vec)] 
    step <- ifelse(length(vec) == 1, 1, vec[2]-vec[1])
    sprintf("%d:%d:%d", start, stop, step)
  }, character(1))
}

# private - fetch and assemble dataset blocks 
# there is a bug
# if blk involves missing data in HDF5, NAs don't come back
# and there is a 0 length list element.
# need a test and fix
multifetch <- function(LL, dataset)  { 
  slicelen <- function(slc)  {
    ss <- as.numeric(strsplit(slc, ':')[[1]])
    r <- (ss[2]-ss[1]) %/% ss[3]
    r+1
  }

  # MM[[d]][[i]] is the length of the ith slice in dimension d
  MM <- lapply(LL, function(L) lapply(L, function(slc) slicelen(slc)))

  # N[d] is the length of the result array in dimension d 
  N <- unlist(lapply(seq_along(MM), function(m) sum(unlist(MM[[m]]))))

  # B[d] is the number of slices in dimension d
  B <- unlist(lapply(seq_along(MM), function(m) length(unlist(MM[[m]]))))

  # pre-allocate result array
  R <- array(rep(0,prod(N)), dim=N) 

  # the total number of fetches is the product of the numbers of slices
  nf <- prod(B)

  # loop over fetches (this can be vapply later)
  for (i in 1:nf)  {

    # select ith block to fetch
    sbs <- rsub4idx(B, i)
    scs <- lapply(seq_along(sbs), function(j) LL[[j]][[sbs[j]]])
  
    # fetch block
    blk <- .getData(dataset, unlist(scs))   # is scs right?

    # put into correct subarray of R 
    nd <- length(LL)
    arglst <- vector(mode="list", length = nd)
    for(d in 1:nd)  {
      umm <- unlist(MM[[d]])
      startpos <- 1
      if (sbs[d] != 1)  {
        startpos <- 1+sum(umm[1:(sbs[d]-1)])
      }
      length <- MM[[d]][[sbs[d]]]
      arglst[[d]] <- seq(startpos, startpos+length-1)
    }  
    R <- do.call('[<-', c(list(R), arglst, list(blk)))
  }
  
  # squash flat dimensions out

  # note: This is kind of a kludge, it would be better to
  # figure out the final dimensions ahead of time and 
  # modify the for loop with conditionals. But if this 
  # works, it evades unnecessary code complexity.

  NN <- N[which(N != 1)]
  R <- array(R, dim=NN)

  R
}

setMethod("show", "HSDSDataset", function(object) {
 cat(paste("rhdf5client HSDSDataset instance, with shape "))
 dput(object@shape)
 cat("  use getData(...) or square brackets to retrieve content.\n")
})


#' bracket method for 1d request from HSDSDataset
#' @param x object of type HSDSDataset
#' @param i vector of indices (first dimension)
#' @param j not used
#' @param \dots not used
#' @param drop logical(1) if TRUE return has no array character
#' @return an array with the elements requested from the HSDSDataset
#' @docType methods
# special case: one-dimensional arrays
setMethod('[', c("HSDSDataset", "numeric", "ANY", "ANY"), 
  function(x, i, j, ..., drop) {
    #.Deprecated("HSDSArray", NULL, deprecate_msg)
    transfermode <- "binary"
    if (x@type$class %in% c("H5T_STRING", "H5T_COMPOUND")) {
      transfermode <- "JSON"
    } 
    getDataList(x, list(i), transfermode=transfermode)
  })

# special case: two-dimensional arrays
#' bracket method for 2d request from HSDSDataset
#' @param x object of type HSDSDataset
#' @param i vector of indices (first dimension)
#' @param j vector of indices (second dimension)
#' @param \dots not used
#' @param drop logical(1) if TRUE return has no array character
#' @return an array with the elements requested from the HSDSDataset
#' @docType methods
setMethod('[', c("HSDSDataset", "numeric", "numeric", "ANY"), 
  function(x, i, j, ..., drop) {
    #.Deprecated("HSDSArray", NULL, deprecate_msg)
    getDataList(x, list(i, j), transfermode='binary')
  })


#Undocumented S4 methods:
#  generic '[' and siglist 'HSDSDataset,numeric,ANY,ANY'
#  generic '[' and siglist 'HSDSDataset,numeric,numeric,ANY'


#' compound operation
#' @importFrom data.table rbindlist as.data.table
#' @param type type
#' @param value value
#' @export
extractCompoundJSON <- function(type, value) {
  stopifnot(type$class == "H5T_COMPOUND")
  
  header <- as.data.table(unlist(lapply(type$fields, function(field) {
    res <- list(switch(field$type$class,
                       "H5T_STRING"=character(0),
                       "H5T_INTEGER"=numeric(0),
                       "H5T_FLOAT"=numeric(0),
                       stop("Unsupported type in H5T_COMPOUND")))
    names(res) <- field$name
    res
  }), recursive = FALSE))
  
  res <- rbindlist(c(list(header), lapply(value, as.list)))
  res
}

---
title: "HSDSArray -- DelayedArray backend for Remote HDF5"
author: "Samuela Pollack and others"
date: "`r format(Sys.time(), '%B %d, %Y')`"
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{HSDSArray DelayedArray backend}
  %\VignetteEncoding{UTF-8}
output:
  BiocStyle::html_document:
    highlight: pygments
    number_sections: yes
    theme: united
    toc: yes
---

```{r setup,echo=FALSE,results="hide"}
suppressPackageStartupMessages({
suppressMessages({
library(rhdf5client)
library(DelayedArray)
})
})
```
# Using the DelayedArray infrastructure

A remote dataset is accessed by giving the URL of the server, the
type of the server (At present the only valid value is `hsds`.), the file domain (path 
to the HDF5 file) and the full path to the dataset inside the HDF5 file. 

## Interface to HSDS (HDF Object Store)
```{r lkdela3}
if (check_hsds()) {
 da <- HSDSArray(URL_hsds(), 'hsds', 
      '/shared/bioconductor/patelGBMSC.h5', '/assay001')
 da
}
```

Again we have DelayedArray capabilities.
```{r lkdel3}
if (check_hsds()) {
 apply(da[,1:4],2,sum)
}
```

---
title: "Extending rhdf5client to cope with more HDF5 details"
author: "Vincent J. Carey, stvjc at channing.harvard.edu"
date: "`r format(Sys.time(), '%B %d, %Y')`"
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Extending rhdf5client to cope with more HDF5 details}
  %\VignetteEncoding{UTF-8}
output:
  BiocStyle::html_document:
    highlight: pygments
    number_sections: yes
    theme: united
    toc: yes
---

```{r setup,echo=FALSE,results="hide"}
suppressPackageStartupMessages({
suppressMessages({
library(rhdf5client)
library(BiocStyle)
})
})
```

# Introduction

Access to HSDS in AWS will end Aug 1 2018.  HDFlab must be
used.  I am focusing for the moment on HDF Server that we
have control over.  HSDS will become open source at the end
of 2018. 

In its current form `r Biocpkg("rhdf5client")` is focused on
providing access to HDF5 datasets corresponding to R matrices.
The test data, provided by HDF Group, for HDF Server illustrates many
aspects of data structure, data type, and server operations,
that rhdf5client does not address.  

We do not have to be comprehensive in our client design but
it would be good to address a few more functionalities.  For
example I do not think we have code that allows management of
a vector as opposed to a matrix.  And our code is not cleanly
factored into server operations, data access, and R interfacing.

```{r basic, echo=FALSE, results="hide"}
library(httr)
library(rjson)

setClass("H5S_dsattrs", representation(attrs="list",
   src="H5S_source", hrefVec="character", theCall="ANY"))
setMethod("show", "H5S_dsattrs", function(object) {
 cat("H5S_dsattrs instance:\n")
 cat("shape.dims:\n ")
 print(object@attrs$shape$dims)
 hrnames = names(object@hrefVec)
 if ("preview" %in% hrnames) cat("A preview URL string is available with prevURL().\n")
})

prevURL = function(x) {
  hr = slot(x, "hrefVec")
  if (!("preview" %in% names(hr))) return(NA)
  hr["preview"]
}

H5S_attr_for_host = function(h5src, pattern, prefix="", postfix="") {
  meta = dsmeta(h5src) # acquires all "host" names in all groups
  dsn = meta[,2] # CharacterList
  ans = lapply(dsn, function(x) grep(pattern, x))
  lens = sapply(ans, length)
  nfi = sum(lens>0)
  if (nfi==0) stop("pattern not found")
  if (nfi>1) {
     message("pattern found in multiple hosts")
     print(grep(pattern, unlist(dsn[which(lens>0)]), value=TRUE))
     message("please refine the pattern used to select the host of interest")
     message("returning NULL")
     return(NULL)
     }
  ind = which(lens>0)
  hostn = dsn[[ind]][ans[[ind]]]
  if (nchar(postfix)>0 | nchar(prefix)>0) hostn=paste0(prefix, hostn, postfix)
  dsreq = sprintf("%s/datasets?%s", .serverURL(h5src), hostn)
  c1 = transl(dsreq)
  dsuuid = c1$datasets
  if (length(dsuuid)!=1) {
   message("there are multiple dataset UUIDs for this request")
   message("returning a list of H5S_dsattrs instances")
   return(lapply(dsuuid, function(x) try(.uuidProc(h5src, hostn, x))))
   }
  .uuidProc(h5src=h5src, hostn=hostn, dsuuid=dsuuid)
}
.uuidProc = function(h5src, hostn, dsuuid) {
  attreq = sprintf("%s/datasets/%s?%s", .serverURL(h5src), dsuuid, hostn)
  a1 = transl(attreq)
  hrefs = sapply(a1$hrefs, "[[", "href")
  names(hrefs) = sapply(a1$hrefs, "[[", "rel")
#  list(atts=a1, src=h5src, theCall=match.call(), hrefs=hrefs)
  new("H5S_dsattrs", attrs=a1, src=h5src, theCall=match.call(),
     hrefVec=hrefs)
}

setGeneric("getSlice", function(h5src, pattern, indxExpr)
   standardGeneric("getSlice"))

setMethod("getSlice", c("H5S_source", "character", "character"),
  function(h5src, pattern, indxExpr) {
   atts = H5S_attr_for_host(h5src, pattern)
   if (is(atts, "list")) {
    message("pattern does not identify unique dataset, returning list of attribute sets")
    return(atts)
    }
   prev = prevURL(atts)
   req = sub("select=..*", paste0("select=", indxExpr), prev)
   transl(req)$value
  })
 
setMethod("getSlice", c("H5S_dsattrs", "missing", "character"),
  function(h5src, pattern, indxExpr) {
   atts = h5src 
   prev = prevURL(atts)
   if (is.na(prev)) {
     if ("data" %in% names(slot(atts, "hrefVec")))
       req = paste0(slot(atts, "hrefVec")["data"], "&select=", 
              indxExpr)
     else stop("no preview URL or 'data' href in this dataset attrs set")
    }
   else req = sub("select=..*", paste0("select=", indxExpr), prev)
   transl(req)$value
  })
 

#> pp[[1]]@attrs$type$base
#[1] "H5T_STD_I16LE"
#> pp[[2]]@attrs$type$base
#[1] "H5T_IEEE_F64BE"
#> pp[[3]]@attrs$type$base
#[1] "H5T_IEEE_F32LE"
```

# A new class that 'simplifies' some additional tasks

This document defines a class that manages dataset attributes
as defined by HDF Server.
```{r lkcl}
getClass("H5S_dsattrs")
```

We can generate an instance of this class for the GTEx data,
and take a relatively unprocessed slice of the content.
```{r dotiss}
tissatt = H5S_attr_for_host( ss, "tissues", 
    prefix="host=", postfix=".h5s.channingremotedata.org")
tissatt
getSlice(tissatt,,"[0:5:1,0:3:1]")
```

This would seem to be a step backwards.  However, we
can get access to more complicated HDF5 data in
this way.

```{r doattdt}
lkdt = H5S_attr_for_host( ss, "datatypes.datasettest.test", prefix="host=", postfix=".h5s.channingremotedata.org")
t(sapply(lkdt, function(x) x@attrs$type))
```

We can also take slices of vectors.

---
title: "rhdf5client -- HDF5 server access"
author: "Vincent J. Carey, stvjc at channing.harvard.edu, Shweta Gopaulakrishnan, reshg at channing.harvard.edu, Samuela Pollack, spollack at jimmy.harvard.edu"
date: "`r format(Sys.time(), '%B %d, %Y')`"
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{rhdf5client -- experiments with interface to remote HDF5}
  %\VignetteEncoding{UTF-8}
output:
  BiocStyle::pdf_document:
    toc: yes
    number_sections: yes
  BiocStyle::html_document:
    highlight: pygments
    number_sections: yes
    theme: united
    toc: yes
---

# HDF5 server 

[HDF Server](https://support.hdfgroup.org/projects/hdfserver/) "extends the HDF5 data model to efficiently store large data objects (e.g. up to multi-TB data arrays) and access them over the web using a RESTful API."  In this package,
several data structures are introduced 

- to model the server data architecture and 
- to perform targeted extraction of numerical data from HDF5 arrays stored on the server. 

We maintain, thanks to a grant from the National Cancer Institute,
the server [http://h5s.channingremotedata.org:5000/](http://h5s.channingremotedata.org:5000/).  Visit this URL to get a flavor of the
server structure: datasets, groups, and datatypes are high-level elements to be manipulated to work
with data values from the server.

A key application of the rhdf5client package is support for the
`r BiocStyle::Biocpkg("restfulSE")` package that defines an interface between
the SummarizedExperiment class and the HDF5 Server.  The server
provides content for `assay()` requests to `RESTfulSummarizedExperiment`
instances.

# Some details

## Motivation

Extensive human and computational effort is expended on
downloading and managing large genomic data at site of analysis.
Interoperable formats that are accessible via generic operations
like those in RESTful APIs may help to improve cost-effectiveness
of genome-scale analyses.

In this report we examine the use of HDF5 server as a back end
for assay data.

A modest server configured to deliver HDF5 content via a
RESTful API has been prepared and is used in this vignette.

## Executive summary

We want to provide rapid read-only access to array-like data.
To do this, the hierarchy and additional formalities of the
HDF5 server data architecture are exposed through R functions
and related classes.  Full details on the HDF5 server
are available at [the HDFgroup site](https://support.hdfgroup.org/projects/hdfserver/).

```{r setup, echo=FALSE}
suppressPackageStartupMessages({
library(rhdf5client)
})
```

The dsmeta function returns top-level groups and datasets. 

```{r dsmeta}
library(rhdf5client)
bigec2 = H5S_source(URL_h5serv())
bigec2
dsmeta(bigec2)[1:2,]      # two groups
dsmeta(bigec2)[1,2][[1]]  # all dataset candidates in group 1
```



## Hierarchy of server resources

### Server

Given the URL of a server running HDF5 server, we create
an instance of `H5S_source`:
```{r doso}
mys = H5S_source(serverURL=URL_h5serv())
mys
```

### Groups

The server identifies a collection of 'groups'.  For the server
we are working with, only one group, at the root, is of interest.
```{r groups}
groups(mys)
```

### Links for a group

There is a class to hold the link set for any group:
```{r links}
lks = links(mys,1)
lks
```

### Dataset access

We use double-bracket subscripting to grab a reference to a
dataset from an H5S source. A dataset must be two-dimensional, 
i.e., accessible with two subscripts.
```{r dataset}
dta = bigec2[["tenx_100k_sorted"]] 
dta
```

### Data

Data are accessed by subscripting. The subscripts are arrays of increasing, 
positive integers.

```{r access}
x = dta[ 15:20, 1905:1906 ]
x
```
(Obsolete) The subscripts are colon-delimited character 
with initial, final and optional stride.

```{r access-obs}
x = dta["15:20", "1904:1906"]
x
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/File.R
\name{HSDSFile}
\alias{HSDSFile}
\title{Construct an object of type HSDSFile}
\usage{
HSDSFile(src, domain)
}
\arguments{
\item{src}{an object of type HSDSSource, the server which exposes the file}

\item{domain}{the domain string; the file's location on the server's
file system.}
}
\value{
an initialized object of type HSDSFile
}
\description{
A HSDSFile is a representation of an HDF5 file the contents of which are accessible 
exposed by a HDF5 server.
}
\details{
This function is deprecated and will be defunct in the next release.
}
\examples{
if (check_hsds()) {
 src <- HSDSSource(URL_hsds())
 f10x <- HSDSFile(src, '/shared/bioconductor/patelGBMSC.h5')
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/File.R
\name{getHSDSFileList}
\alias{getHSDSFileList}
\title{List files on a directory on HSDS-server}
\usage{
getHSDSFileList(url, directory = NULL)
}
\arguments{
\item{url, }{containing url of the server and root domain.}

\item{directory, }{containing name of the directory}
}
\value{
List of all HDF5-files on the server or all files of the collection
}
\description{
This funstions returns a list of all HDF5-files on the server or all files of the collection
}
\examples{
if (check_hsds()) {
 url <- 'https://alserglab.wustl.edu/hsds/?domain=/counts'
 getHSDSFileList(url)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/File.R
\docType{class}
\name{HSDSFile-class}
\alias{HSDSFile-class}
\title{An S4 class to represent an HDF5 file accessible from a server.}
\description{
An S4 class to represent an HDF5 file accessible from a server.
}
\section{Slots}{

\describe{
\item{\code{HSDSSource}}{an object of type HSDSSource}

\item{\code{domain}}{the file's domain on the server; more or less, an alias for its 
location in the external server file system}

\item{\code{dsetdf}}{a data.frame that caches often-used information about the file}

\item{\code{type}}{string representing type of the file: "domain" for file and "folder" for directory}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RHDF5Array.R
\docType{class}
\name{HSDSMatrix}
\alias{HSDSMatrix}
\alias{HSDSMatrix-class}
\title{DelayedMatrix subclass for a two-dimensional HSDSArray}
\description{
DelayedMatrix subclass for a two-dimensional HSDSArray
}
\seealso{
Other HSDSArray: 
\code{\link{HSDSArray}},
\code{\link{as}()}
}
\concept{HSDSArray}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RHDF5Array.R
\name{extract_array}
\alias{extract_array}
\alias{extract_array,HSDSArraySeed-method}
\title{Access dataset backed by an HSDSArraySeed}
\usage{
\S4method{extract_array}{HSDSArraySeed}(x, index)
}
\arguments{
\item{x}{An object of type HSDSArraySeed}

\item{index}{A list of numeric vectors to be accessed, one vector 
for each dimension of the array object. A NULL vector indicates
the entire range of indices in that dimension. A zero-length
vector indicates no indices in the relevant dimension. (Accordingly,
any zero-length vector of indices will result in an empty array
being returned.)}
}
\value{
An array containing the data elements corresponding to the 
indices requested
}
\description{
Access dataset backed by an HSDSArraySeed
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Source.R
\docType{class}
\name{HSDSSource-class}
\alias{HSDSSource-class}
\title{An S4 class to represent a HDF5 server listening on a port.}
\description{
This class is deprecated and will be defunct in the next release.
}
\section{Slots}{

\describe{
\item{\code{endpoint}}{URL for server}

\item{\code{type}}{Type of server software at the source; must be 
either 'h5serv' or (default) 'hsds'}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RHDF5Array.R
\docType{class}
\name{HSDSArraySeed}
\alias{HSDSArraySeed}
\alias{HSDSArraySeed-class}
\title{HSDSArraySeed for HSDSArray backend to DelayedArray}
\usage{
HSDSArraySeed(endpoint, svrtype, domain, dsetname)
}
\arguments{
\item{endpoint}{URL of remote server}

\item{svrtype}{type of server, must be either 'hsds' or 'h5serv'}

\item{domain}{HDF5 domain of H5 file on server}

\item{dsetname}{complete internal path to dataset in H5 file}
}
\value{
An initialized object of type HSDSArraySeed
}
\description{
HSDSArraySeed for HSDSArray backend to DelayedArray

Construct an object of type HSDSArraySeed
}
\section{Slots}{

\describe{
\item{\code{endpoint}}{URL of remote server}

\item{\code{svrtype}}{type of server, must be either 'hsds' or 'h5serv'}

\item{\code{domain}}{HDF5 domain of H5 file on server}

\item{\code{dsetname}}{complete internal path to dataset in H5 file}

\item{\code{dataset}}{object of type HSDSDataset for access to the H5 dataset}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RHDF5Array.R
\name{check_hsds}
\alias{check_hsds}
\title{a test request}
\usage{
check_hsds()
}
\value{
logical, TRUE if hsds behaving as expected
}
\description{
a test request
}
\examples{
check_hsds()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Dataset.R
\docType{methods}
\name{getData}
\alias{getData}
\alias{getData,HSDSDataset,character,character-method}
\alias{getData,HSDSDataset,character,missing-method}
\alias{getData,HSDSDataset,list,character-method}
\alias{getData,HSDSDataset,list,missing-method}
\alias{getData,HSDSDataset,list,,missing-method}
\title{extract elements of a one or two-dimensional HSDSDataset}
\usage{
getData(dataset, indices, transfermode)

\S4method{getData}{HSDSDataset,character,character}(dataset, indices, transfermode)

\S4method{getData}{HSDSDataset,character,missing}(dataset, indices)

\S4method{getData}{HSDSDataset,list,character}(dataset, indices, transfermode)

\S4method{getData}{HSDSDataset,list,missing}(dataset, indices)
}
\arguments{
\item{dataset}{An object of type HSDSDataset, the dataset to access.}

\item{indices}{The indices of the data to fetch}

\item{transfermode}{Either 'JSON' or 'binary' (default)}
}
\value{
an Array containing the data fetched from the server
}
\description{
Fetch data from a remote dataset
}
\details{
The servers require data to be fetched in slices, i.e., in sets of 
for which the indices of each dimension are of the form start:stop:step.
More complex sets of indices will be split into slices and fetched in
multiple requests. This is opaque to the user, but may enter into 
considerations of data access patterns, e.g., for performance-tuning.
}
\examples{
if (check_hsds()) {
 s <- HSDSSource(URL_hsds())
 f <- HSDSFile(s, '/shared/bioconductor/patelGBMSC.h5')
 d <- HSDSDataset(f, '/assay001')
 x <- getData(d, c('1:4', '1:27998'), transfermode='JSON')
 xb <- getData(d, c('1:4', '1:27998'), transfermode='binary')
 # x <- getData(d, c(1:4, 1:27998), transfermode='JSON') # method missing?
 x 
 xb
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Source.R
\docType{methods}
\name{listDomains}
\alias{listDomains}
\alias{listDomains,HSDSSource,character-method}
\alias{listDomains,HSDSSource,missing-method}
\title{List files and subdirectories of a domain}
\usage{
listDomains(object, rootdir)

\S4method{listDomains}{HSDSSource,character}(object, rootdir)

\S4method{listDomains}{HSDSSource,missing}(object)
}
\arguments{
\item{object}{An object of type HSDSSource}

\item{rootdir}{A slash-separated directory in the HSDSSource file system.}
}
\value{
a vector of domains in the rootdir
}
\description{
The user needs to give the domain to start in. The search
will be non-recursive. I.e., output for domain '/home/jreadey/' will 
not return the files in '/home/jreadey/HDFLabTutorial/'
}
\details{
This function is deprecated and will be defunct in the next release.
}
\examples{
src.hsds <- HSDSSource(URL_hsds())
listDomains(src.hsds, '/shared')
}

\name{rhdf5client-deprecated}
\alias{rhdf5client-deprecated}
\title{Deprecated functions in package \sQuote{rhdf5client}}

\description{
  These functions are provided for compatibility with older versions
  of \sQuote{rhdf5client} only, and will be defunct at the next release.
}

\details{

  The following functions are deprecated and will be made defunct in the next release: 
  \itemize{
    \item{URL_h5serv} 
    \item{URL_hsds}
    \item{dsmeta}
    \item{getReq}
    \item{groups}
    \item{setPath}
    \item{links}
    \item{transfermode}
    \item{dataset}
    \item{internalDim}
    \item{hsdsInfo}
    \item{domains}
    \item{getDatasetUUIDs}
    \item{getDatasetAttrs}
    \item{getDims}
    \item{getHRDF}
    \item{H5S_dataset2}
    \item{getDatasetSlice}
    \item{fetchDatasets}
    \item{isplit}
    \item{sproc}
    \item{listDomains}
    \item{listDatasets}
    \item{getData}
% how should these be declared?
%   \item{subscript}
%   \item{double-subscript}
%   \item{transfermode<-}
  }
  The following classes are deprecated and will be made defunct in the next release: 
  \itemize{
    \item{H5S_source}
    \item{H5S_dataset}
    \item{H5S_Array}
    \item{H5S_Matrix}
    \item{HSDSSource}
    \item{HSDSFile}
    \item{HSDSDataset}
  }
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/URL_hsds.R
\name{URL_hsds}
\alias{URL_hsds}
\title{manage hsds URL}
\usage{
URL_hsds()
}
\value{
URL of hsds server
}
\description{
manage hsds URL
}
\examples{
URL_hsds()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Dataset.R
\docType{class}
\name{HSDSDataset-class}
\alias{HSDSDataset-class}
\title{An S4 class to represent a dataset in a HDF5 file.}
\description{
An S4 class to represent a dataset in a HDF5 file.
}
\section{Slots}{

\describe{
\item{\code{file}}{An object of type HSDSFile; the file in which the dataset is resident.}

\item{\code{path}}{The dataset's path in the internal HDF5 hiearchy.}

\item{\code{uuid}}{The unique unit ID by which the dataset is accessed in the server 
database system.}

\item{\code{shape}}{The dimensions of the dataset}

\item{\code{type}}{The dataset's HDF5 datatype}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RHDF5Array.R
\docType{class}
\name{HSDSArray}
\alias{HSDSArray}
\alias{HSDSArray-class}
\title{A DelayedArray backend for accessing a remote HDF5 server.}
\usage{
HSDSArray(endpoint, svrtype, domain, dsetname)
}
\arguments{
\item{endpoint}{URL of remote server}

\item{svrtype}{type of server, must be either 'hsds' or 'h5serv'}

\item{domain}{HDF5 domain of H5 file on server}

\item{dsetname}{complete internal path to dataset in H5 file}
}
\value{
An initialized object of type HSDSArray
}
\description{
A DelayedArray backend for accessing a remote HDF5 server.

Construct an object of type HSDSArray directly from the data
members of its seed
}
\examples{
if (check_hsds()) {
 HSDSArray(URL_hsds(), 
    "hsds", "/shared/bioconductor/patelGBMSC.h5", "/assay001")
}
}
\seealso{
Other HSDSArray: 
\code{\link{HSDSMatrix}},
\code{\link{as}()}
}
\concept{HSDSArray}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Dataset.R
\name{HSDSDataset}
\alias{HSDSDataset}
\title{Construct an object of type HSDSDataset 
A HSDSDataset is a representation of a dataset in a HDF5 file.}
\usage{
HSDSDataset(file, path)
}
\arguments{
\item{file}{An object of type HSDSFile which hosts the dataset}

\item{path}{The complete intrafile path to the dataset}
}
\value{
An initialized object of type HSDSDataset
}
\description{
Construct an object of type HSDSDataset 
A HSDSDataset is a representation of a dataset in a HDF5 file.
}
\examples{
if (check_hsds()) {
 src <- HSDSSource(URL_hsds())
 f <- HSDSFile(src, '/shared/bioconductor/patelGBMSC.h5')
 d <- HSDSDataset(f, '/assay001')
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RHDF5Array.R
\name{dim}
\alias{dim}
\alias{dim,HSDSArraySeed-method}
\title{Obtain dimensions of an object of type HSDSArraySeed}
\usage{
\S4method{dim}{HSDSArraySeed}(x)
}
\arguments{
\item{x}{An object of type HSDSArraySeed}
}
\value{
A numeric vector of the dimensions
}
\description{
(required by DelayedArray seed contract)
HDF server content is assumed transposed relative to R matrix layout.
This anticipates H5 datasets on the server with rows for 
experimental samples and columns for *-omic features. The 
Bioconductor SummarizedExperiment requires *-omic features in 
rows and samples in columns.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/indx.R
\name{isplit}
\alias{isplit}
\alias{sproc}
\title{This function is deprecated and will be defunct in the next release.}
\usage{
isplit(x)

sproc(spl)
}
\arguments{
\item{x}{a numeric vector (should be integers)}

\item{spl}{output of isplit}
}
\value{
list of vectors of integers which can be expressed as initial/final/stride triplets

list of colon-delimited strings each with initial/final/stride triplet
}
\description{
isplit converts a numeric vector into a list of sequences for compact reexpression
}
\examples{
inds = c(1:10, seq(25,50,2), seq(200,150,-2))
sproc(isplit(inds))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Source.R
\name{HSDSSource}
\alias{HSDSSource}
\title{Construct an object of type HSDSSource.}
\usage{
HSDSSource(endpoint, type = "hsds")
}
\arguments{
\item{endpoint}{URL for server}

\item{type}{Type of server software at the source; must be}
}
\value{
An object of type HSDSSource
}
\description{
A HSDSSource is a representation of a URL which provides access to a HDF5 
server (either h5serv or hsds.)
}
\details{
This function is deprecated and will be defunct in the next release.
}
\examples{
if (check_hsds()) {
 src.hsds <- HSDSSource(URL_hsds())
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RHDF5Array.R
\name{as}
\alias{as}
\title{Coercion method from HSDSMatrix to its superclass HSDSArray}
\description{
Coercion method from HSDSMatrix to its superclass HSDSArray
}
\seealso{
Other HSDSArray: 
\code{\link{HSDSArray}},
\code{\link{HSDSMatrix}}
}
\concept{HSDSArray}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rhdf5client2.R
\docType{package}
\name{rhdf5client}
\alias{rhdf5client-package}
\alias{rhdf5client}
\title{rhdf5client: A package for accessing HDFGroup HDF5 servers from R.}
\description{
The rhdf5client package provides read-only access to HDF5 files
maintained on a server. The HDFGroup provides two servers, an
obsolescent one called `h5serv` and the newer prototype called `hsds`.
}
\author{
\strong{Maintainer}: Vincent Carey \email{stvjc@channing.harvard.edu}

Authors:
\itemize{
  \item Samuela Pollack \email{spollack@jimmy.harvard.edu}
  \item Shweta Gopaulakrishnan \email{reshg@channing.harvard.edu}
  \item BJ Stubbs \email{rebjh@channing.harvard.edu}
  \item Alexey Sergushichev \email{alsergbox@gmail.com}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/File.R
\name{listDatasets}
\alias{listDatasets}
\title{Search inner file hierarchy for datasets}
\usage{
listDatasets(file)
}
\arguments{
\item{file}{an object of type HSDSFile to be searched}
}
\value{
a list of inner-paths
}
\description{
The datasets in an HDF5 file are organized internally by groups.
This routine traverses the internal group hiearchy, locates
all datasets and prints a list of them. Note that if the 
file's group hiearchy is complex, this could be time-consuming.
}
\details{
This function is deprecated and will be defunct in the next release.
}
\examples{
if (check_hsds()) {
 src <- HSDSSource(URL_hsds())
 f <- HSDSFile(src, '/shared/bioconductor/patelGBMSC.h5')
 listDatasets(f)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Dataset.R
\docType{methods}
\name{[,HSDSDataset,numeric,ANY,ANY-method}
\alias{[,HSDSDataset,numeric,ANY,ANY-method}
\title{bracket method for 1d request from HSDSDataset}
\usage{
\S4method{[}{HSDSDataset,numeric,ANY,ANY}(x, i, j, ..., drop = TRUE)
}
\arguments{
\item{x}{object of type HSDSDataset}

\item{i}{vector of indices (first dimension)}

\item{j}{not used}

\item{\dots}{not used}

\item{drop}{logical(1) if TRUE return has no array character}
}
\value{
an array with the elements requested from the HSDSDataset
}
\description{
bracket method for 1d request from HSDSDataset
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Dataset.R
\name{extractCompoundJSON}
\alias{extractCompoundJSON}
\title{compound operation}
\usage{
extractCompoundJSON(type, value)
}
\arguments{
\item{type}{type}

\item{value}{value}
}
\description{
compound operation
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RHDF5Array.R
\name{dimnames}
\alias{dimnames}
\alias{dimnames,HSDSArraySeed-method}
\title{Obtain names of dimensions for an object of type HSDSArraySeed}
\usage{
\S4method{dimnames}{HSDSArraySeed}(x)
}
\arguments{
\item{x}{An object of type HSDSArraySeed}
}
\value{
A NULL list of length equal to the array dimensionality
}
\description{
(required by DelayedArray seed contract, returns NULL list)
}
